@@ -5,7 +5,7 @@
  *
  * The MIT License
  *
- * @copyright Copyright (c) 2017-2018 TileDB Inc.
+ * @copyright Copyright (c) 2017-2021 TileDB Inc.
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
  * of this software and associated documentation files (the "Software"), to deal
@@ -31,8 +31,8 @@
  */
 
 #include "catch.hpp"
+#include "test/src/helpers.h"
 #include "tiledb/sm/c_api/tiledb.h"
-#include "tiledb/sm/misc/stats.h"
 #include "tiledb/sm/misc/utils.h"
 #ifdef _WIN32
 #include "tiledb/sm/filesystem/win.h"
@@ -44,11 +44,17 @@
 #include <sstream>
 #include <thread>
 
+using namespace tiledb::test;
+
 struct VFSFx {
   const std::string HDFS_TEMP_DIR = "hdfs://localhost:9000/tiledb_test/";
   const std::string S3_PREFIX = "s3://";
-  const std::string S3_BUCKET = S3_PREFIX + random_bucket_name("tiledb") + "/";
+  const std::string S3_BUCKET = S3_PREFIX + random_name("tiledb") + "/";
   const std::string S3_TEMP_DIR = S3_BUCKET + "tiledb_test/";
+  const std::string AZURE_PREFIX = "azure://";
+  const std::string AZURE_CONTAINER =
+      AZURE_PREFIX + random_name("tiledb") + "/";
+  const std::string AZURE_TEMP_DIR = AZURE_CONTAINER + "tiledb_test/";
 #ifdef _WIN32
   const std::string FILE_TEMP_DIR =
       tiledb::sm::Win::current_dir() + "\\tiledb_test\\";
@@ -57,6 +63,7 @@ struct VFSFx {
                                     tiledb::sm::Posix::current_dir() +
                                     "/tiledb_test/";
 #endif
+  const std::string MEMFS_TEMP_DIR = std::string("mem://tiledb_test/");
 
   // TileDB context and vfs
   tiledb_ctx_t* ctx_;
@@ -65,6 +72,8 @@ struct VFSFx {
   // Supported filesystems
   bool supports_s3_;
   bool supports_hdfs_;
+  bool supports_azure_;
+  bool supports_memfs_;
 
   // Functions
   VFSFx();
@@ -72,10 +81,11 @@ struct VFSFx {
   void check_vfs(const std::string& path);
   void check_write(const std::string& path);
   void check_move(const std::string& path);
+  void check_copy(const std::string& path);
   void check_read(const std::string& path);
   void check_append(const std::string& path);
   void check_ls(const std::string& path);
-  static std::string random_bucket_name(const std::string& prefix);
+  static std::string random_name(const std::string& prefix);
   void set_supported_fs();
   void set_num_vfs_threads(unsigned num_threads);
 };
@@ -100,13 +110,9 @@ void VFSFx::set_supported_fs() {
   tiledb_ctx_t* ctx = nullptr;
   REQUIRE(tiledb_ctx_alloc(nullptr, &ctx) == TILEDB_OK);
 
-  int is_supported = 0;
-  int rc = tiledb_ctx_is_supported_fs(ctx, TILEDB_S3, &is_supported);
-  REQUIRE(rc == TILEDB_OK);
-  supports_s3_ = (bool)is_supported;
-  rc = tiledb_ctx_is_supported_fs(ctx, TILEDB_HDFS, &is_supported);
-  REQUIRE(rc == TILEDB_OK);
-  supports_hdfs_ = (bool)is_supported;
+  bool supports_gcs;  // unused
+  get_supported_fs(
+      &supports_s3_, &supports_hdfs_, &supports_azure_, &supports_gcs);
 
   tiledb_ctx_free(&ctx);
 }
@@ -141,14 +147,32 @@ void VFSFx::set_num_vfs_threads(unsigned num_threads) {
     REQUIRE(error == nullptr);
 #endif
   }
+  if (supports_azure_) {
+    REQUIRE(
+        tiledb_config_set(
+            config,
+            "vfs.azure.storage_account_name",
+            "devstoreaccount1",
+            &error) == TILEDB_OK);
+    REQUIRE(
+        tiledb_config_set(
+            config,
+            "vfs.azure.storage_account_key",
+            "Eby8vdM02xNOcqFlqUwJPLlmEtlCDXJ1OUzFT50uSRZ6IFsuFq2UVErCz4I6tq/"
+            "K1SZFPTOtr/KBHBeksoGMGw==",
+            &error) == TILEDB_OK);
+    REQUIRE(
+        tiledb_config_set(
+            config,
+            "vfs.azure.blob_endpoint",
+            "127.0.0.1:10000/devstoreaccount1",
+            &error) == TILEDB_OK);
+    REQUIRE(
+        tiledb_config_set(config, "vfs.azure.use_https", "false", &error) ==
+        TILEDB_OK);
+  }
 
   // Set number of threads across all backends.
-  REQUIRE(
-      tiledb_config_set(
-          config,
-          "vfs.num_threads",
-          std::to_string(num_threads).c_str(),
-          &error) == TILEDB_OK);
   REQUIRE(
       tiledb_config_set(
           config,
@@ -285,6 +309,9 @@ void VFSFx::check_vfs(const std::string& path) {
   REQUIRE(is_file);
   rc = tiledb_vfs_remove_file(ctx_, vfs_, foo_file.c_str());
   REQUIRE(rc == TILEDB_OK);
+  rc = tiledb_vfs_is_file(ctx_, vfs_, foo_file.c_str(), &is_file);
+  REQUIRE(rc == TILEDB_OK);
+  REQUIRE(!is_file);
 
   // Check write and append
   check_write(path);
@@ -296,6 +323,14 @@ void VFSFx::check_vfs(const std::string& path) {
   // Move
   check_move(path);
 
+#ifndef _WIN32
+  // Copy
+  if (path != MEMFS_TEMP_DIR) {
+    // copy not yet supported for memfs
+    check_copy(path);
+  }
+#endif
+
   // Ls
   check_ls(path);
 
@@ -417,7 +452,7 @@ void VFSFx::check_move(const std::string& path) {
   // Move from one bucket to another (only for S3)
   if (supports_s3_) {
     if (path == S3_TEMP_DIR) {
-      std::string bucket2 = S3_PREFIX + random_bucket_name("tiledb") + "/";
+      std::string bucket2 = S3_PREFIX + random_name("tiledb") + "/";
       std::string subdir3 = bucket2 + "tiledb_test/subdir3/";
       std::string file3 = subdir3 + "file2";
       int is_bucket = 0;
@@ -444,6 +479,117 @@ void VFSFx::check_move(const std::string& path) {
   }
 }
 
+#ifndef _WIN32
+void VFSFx::check_copy(const std::string& path) {
+  // Do not support HDFS
+  if (supports_hdfs_)
+    return;
+
+  // Copy file
+  auto file = path + "file";
+  auto file2 = path + "file2";
+  int is_file = 0;
+  int rc = tiledb_vfs_touch(ctx_, vfs_, file.c_str());
+  REQUIRE(rc == TILEDB_OK);
+  rc = tiledb_vfs_is_file(ctx_, vfs_, file.c_str(), &is_file);
+  REQUIRE(rc == TILEDB_OK);
+  REQUIRE(is_file);
+  rc = tiledb_vfs_copy_file(ctx_, vfs_, file.c_str(), file2.c_str());
+  REQUIRE(rc == TILEDB_OK);
+  rc = tiledb_vfs_is_file(ctx_, vfs_, file.c_str(), &is_file);
+  REQUIRE(rc == TILEDB_OK);
+  REQUIRE(is_file);
+  rc = tiledb_vfs_is_file(ctx_, vfs_, file2.c_str(), &is_file);
+  REQUIRE(rc == TILEDB_OK);
+  REQUIRE(is_file);
+
+  // Move directory with subdirectories and files
+  auto dir = path + "dir/";
+  auto dir2 = path + "dir2/";
+  auto subdir = path + "dir/subdir/";
+  auto subdir2 = path + "dir2/subdir/";
+  file = dir + "file";
+  file2 = subdir + "file2";
+  auto new_file = dir2 + "file";
+  auto new_file2 = subdir2 + "file2";
+  int is_dir = 0;
+  rc = tiledb_vfs_create_dir(ctx_, vfs_, dir.c_str());
+  REQUIRE(rc == TILEDB_OK);
+  rc = tiledb_vfs_is_dir(ctx_, vfs_, dir.c_str(), &is_dir);
+  REQUIRE(rc == TILEDB_OK);
+  if (path == S3_TEMP_DIR)
+    REQUIRE(!is_dir);  // No empty dirs exist in S3
+  else
+    REQUIRE(is_dir);
+  rc = tiledb_vfs_create_dir(ctx_, vfs_, subdir.c_str());
+  REQUIRE(rc == TILEDB_OK);
+  rc = tiledb_vfs_is_dir(ctx_, vfs_, subdir.c_str(), &is_dir);
+  REQUIRE(rc == TILEDB_OK);
+  if (path == S3_TEMP_DIR)
+    REQUIRE(!is_dir);  // No empty dirs exist in S3
+  else
+    REQUIRE(is_dir);
+  rc = tiledb_vfs_touch(ctx_, vfs_, file.c_str());
+  REQUIRE(rc == TILEDB_OK);
+  rc = tiledb_vfs_is_file(ctx_, vfs_, file.c_str(), &is_file);
+  REQUIRE(rc == TILEDB_OK);
+  REQUIRE(is_file);
+  rc = tiledb_vfs_touch(ctx_, vfs_, file2.c_str());
+  REQUIRE(rc == TILEDB_OK);
+  rc = tiledb_vfs_is_file(ctx_, vfs_, file2.c_str(), &is_file);
+  REQUIRE(rc == TILEDB_OK);
+  REQUIRE(is_file);
+
+  rc = tiledb_vfs_is_dir(ctx_, vfs_, dir2.c_str(), &is_dir);
+  REQUIRE(rc == TILEDB_OK);
+  if (is_dir) {
+    rc = tiledb_vfs_remove_dir(ctx_, vfs_, dir2.c_str());
+    REQUIRE(rc == TILEDB_OK);
+  }
+  rc = tiledb_vfs_copy_dir(ctx_, vfs_, dir.c_str(), dir2.c_str());
+  REQUIRE(rc == TILEDB_OK);
+  rc = tiledb_vfs_is_dir(ctx_, vfs_, dir2.c_str(), &is_dir);
+  REQUIRE(rc == TILEDB_OK);
+  REQUIRE(is_dir);
+  rc = tiledb_vfs_is_dir(ctx_, vfs_, subdir2.c_str(), &is_dir);
+  REQUIRE(rc == TILEDB_OK);
+  REQUIRE(is_dir);
+  rc = tiledb_vfs_is_file(ctx_, vfs_, new_file.c_str(), &is_file);
+  REQUIRE(rc == TILEDB_OK);
+  REQUIRE(is_file);
+  rc = tiledb_vfs_is_file(ctx_, vfs_, new_file2.c_str(), &is_file);
+  REQUIRE(rc == TILEDB_OK);
+  REQUIRE(is_file);
+
+  // Copy from one bucket to another (only for S3)
+  if (supports_s3_) {
+    std::string bucket2 = S3_PREFIX + random_name("tiledb") + "/";
+    std::string subdir3 = bucket2 + "tiledb_test/subdir3/";
+    std::string file3 = subdir3 + "file2";
+    int is_bucket = 0;
+
+    rc = tiledb_vfs_is_bucket(ctx_, vfs_, bucket2.c_str(), &is_bucket);
+    REQUIRE(rc == TILEDB_OK);
+    if (is_bucket) {
+      rc = tiledb_vfs_remove_bucket(ctx_, vfs_, bucket2.c_str());
+      REQUIRE(rc == TILEDB_OK);
+    }
+
+    rc = tiledb_vfs_create_bucket(ctx_, vfs_, bucket2.c_str());
+    REQUIRE(rc == TILEDB_OK);
+
+    rc = tiledb_vfs_copy_dir(ctx_, vfs_, subdir2.c_str(), subdir3.c_str());
+    REQUIRE(rc == TILEDB_OK);
+    rc = tiledb_vfs_is_file(ctx_, vfs_, file3.c_str(), &is_file);
+    REQUIRE(rc == TILEDB_OK);
+    REQUIRE(is_file);
+
+    rc = tiledb_vfs_remove_bucket(ctx_, vfs_, bucket2.c_str());
+    REQUIRE(rc == TILEDB_OK);
+  }
+}
+#endif
+
 void VFSFx::check_write(const std::string& path) {
   // File write and file size
   int is_file = 0;
@@ -749,14 +895,14 @@ void VFSFx::check_ls(const std::string& path) {
   CHECK(children[3] == subdir2);
 }
 
-std::string VFSFx::random_bucket_name(const std::string& prefix) {
+std::string VFSFx::random_name(const std::string& prefix) {
   std::stringstream ss;
   ss << prefix << "-" << std::this_thread::get_id() << "-"
      << TILEDB_TIMESTAMP_NOW_MS;
   return ss.str();
 }
 
-TEST_CASE_METHOD(VFSFx, "C API: Test virtual filesystem", "[capi], [vfs]") {
+TEST_CASE_METHOD(VFSFx, "C API: Test virtual filesystem", "[capi][vfs]") {
   tiledb_stats_enable();
   tiledb_stats_reset();
 
@@ -764,19 +910,18 @@ TEST_CASE_METHOD(VFSFx, "C API: Test virtual filesystem", "[capi], [vfs]") {
     check_vfs(S3_TEMP_DIR);
   else if (supports_hdfs_)
     check_vfs(HDFS_TEMP_DIR);
-  else
+  else {
     check_vfs(FILE_TEMP_DIR);
-
-  CHECK(tiledb::sm::stats::all_stats.counter_vfs_read_num_parallelized == 0);
-  CHECK(
-      tiledb::sm::stats::all_stats.counter_vfs_posix_write_num_parallelized ==
-      0);
+    if (supports_memfs_) {
+      check_vfs(MEMFS_TEMP_DIR);
+    }
+  }
 }
 
 TEST_CASE_METHOD(
     VFSFx,
     "C API: Test virtual filesystem when S3 is not supported",
-    "[capi], [vfs]") {
+    "[capi][vfs]") {
   if (!supports_s3_) {
     tiledb_vfs_t* vfs;
     int rc = tiledb_vfs_alloc(ctx_, nullptr, &vfs);
@@ -788,7 +933,7 @@ TEST_CASE_METHOD(
 }
 
 TEST_CASE_METHOD(
-    VFSFx, "C API: Test virtual filesystem config", "[capi], [vfs]") {
+    VFSFx, "C API: Test virtual filesystem config", "[capi][vfs]") {
   // Prepare a config
   tiledb_error_t* error = nullptr;
   tiledb_config_t* config;
@@ -822,28 +967,19 @@ TEST_CASE_METHOD(
   tiledb_vfs_free(&vfs);
 }
 
-TEST_CASE_METHOD(VFSFx, "C API: Test VFS parallel I/O", "[capi], [vfs]") {
+TEST_CASE_METHOD(VFSFx, "C API: Test VFS parallel I/O", "[capi][vfs]") {
   tiledb_stats_enable();
   tiledb_stats_reset();
   set_num_vfs_threads(4);
 
   if (supports_s3_) {
     check_vfs(S3_TEMP_DIR);
-    CHECK(tiledb::sm::stats::all_stats.counter_vfs_read_num_parallelized > 0);
   } else if (supports_hdfs_) {
     check_vfs(HDFS_TEMP_DIR);
-    CHECK(tiledb::sm::stats::all_stats.counter_vfs_read_num_parallelized == 0);
   } else {
     check_vfs(FILE_TEMP_DIR);
-    CHECK(tiledb::sm::stats::all_stats.counter_vfs_read_num_parallelized > 0);
-#ifdef _WIN32
-    CHECK(
-        tiledb::sm::stats::all_stats.counter_vfs_win32_write_num_parallelized >
-        0);
-#else
-    CHECK(
-        tiledb::sm::stats::all_stats.counter_vfs_posix_write_num_parallelized >
-        0);
-#endif
+    if (supports_memfs_) {
+      check_vfs(MEMFS_TEMP_DIR);
+    }
   }
 }
