@@ -5,7 +5,7 @@
  *
  * The MIT License
  *
- * @copyright Copyright (c) 2017-2019 TileDB, Inc.
+ * @copyright Copyright (c) 2017-2021 TileDB, Inc.
  * @copyright Copyright (c) 2016 MIT and Intel Corporation
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
@@ -31,32 +31,34 @@
  * This file defines class ArraySchema.
  */
 
-#ifndef TILEDB_ARRAY_METADATA_H
-#define TILEDB_ARRAY_METADATA_H
+#ifndef TILEDB_ARRAY_SCHEMA_H
+#define TILEDB_ARRAY_SCHEMA_H
 
-#define __STDC_FORMAT_MACROS
-
-#include <string>
-#include <typeinfo>
 #include <unordered_map>
-#include <vector>
 
-#include "tiledb/sm/array_schema/attribute.h"
-#include "tiledb/sm/array_schema/dimension.h"
-#include "tiledb/sm/array_schema/domain.h"
-#include "tiledb/sm/buffer/buffer.h"
-#include "tiledb/sm/enums/array_type.h"
-#include "tiledb/sm/enums/compressor.h"
-#include "tiledb/sm/enums/datatype.h"
-#include "tiledb/sm/enums/layout.h"
+#include "tiledb/common/status.h"
 #include "tiledb/sm/filter/filter_pipeline.h"
 #include "tiledb/sm/misc/constants.h"
-#include "tiledb/sm/misc/status.h"
+#include "tiledb/sm/misc/hilbert.h"
 #include "tiledb/sm/misc/uri.h"
+#include "tiledb/sm/misc/uuid.h"
+
+using namespace tiledb::common;
 
 namespace tiledb {
 namespace sm {
 
+class Attribute;
+class Buffer;
+class ConstBuffer;
+class Dimension;
+class Domain;
+
+enum class ArrayType : uint8_t;
+enum class Compressor : uint8_t;
+enum class Datatype : uint8_t;
+enum class Layout : uint8_t;
+
 /** Specifies the array schema. */
 class ArraySchema {
  public:
@@ -84,46 +86,29 @@ class ArraySchema {
   /*               API                 */
   /* ********************************* */
 
+  /**
+   * Returns true if the array allows coordinate duplicates. Applicable
+   * only to sparse arrays, dense arrays do not allow duplicates.
+   */
+  bool allows_dups() const;
+
   /** Returns the array type. */
   ArrayType array_type() const;
 
   /** Returns the array uri. */
   const URI& array_uri() const;
 
-  /** Returns a constant pointer to the selected attribute (NULL if error). */
+  /**
+   * Returns a constant pointer to the selected attribute (nullptr if it
+   * does not exist).
+   */
   const Attribute* attribute(unsigned int id) const;
 
-  /** Returns a constant pointer to the selected attribute (NULL if error). */
-  const Attribute* attribute(std::string name) const;
-
   /**
-   * Returns the given attribute name as it would be stored in the schema. E.g.
-   * if the argument is "" (empty string), this returns the default anonymous
-   * attribute name, which is what is stored in the schema for anonymous
-   * attributes.
-   *
-   * @param attribute Attribute name
-   * @param normalized_name Will hold the normalized name
-   * @return Status
+   * Returns a constant pointer to the selected attribute (nullptr if it
+   * does not exist).
    */
-  static Status attribute_name_normalized(
-      const char* attribute, std::string* normalized_name);
-
-  /**
-   * Returns the given attribute names as they would be stored in the schema.
-   * E.g. if an input name is "" (empty string), this returns it as the default
-   * anonymous attribute name, which is what is stored in the schema for
-   * anonymous attributes.
-   *
-   * @param attributes Attribute names to normalize
-   * @param num_attributes Number of attribute names
-   * @param normalized_names Will hold the normalized names
-   * @return Status
-   */
-  static Status attribute_names_normalized(
-      const char** attributes,
-      unsigned num_attributes,
-      std::vector<std::string>* normalized_names);
+  const Attribute* attribute(const std::string& name) const;
 
   /** Returns the number of attributes. */
   unsigned int attribute_num() const;
@@ -137,16 +122,17 @@ class ArraySchema {
   /** Returns the cell order. */
   Layout cell_order() const;
 
-  /** Returns the size of cell on the input attribute. */
-  uint64_t cell_size(const std::string& attribute) const;
+  /** Returns the size of cell on the input attribute/dimension. */
+  uint64_t cell_size(const std::string& name) const;
 
-  /** Returns the number of values per cell of the input attribute. */
-  unsigned int cell_val_num(const std::string& attribute) const;
+  /** Returns the number of values per cell of the input attribute/dimension. */
+  unsigned int cell_val_num(const std::string& name) const;
 
-  /**
-   * Return a pointer to the pipeline used for offsets of variable-sized cells.
-   */
-  const FilterPipeline* cell_var_offsets_filters() const;
+  /** Return the filter pipeline used for offsets of variable-sized cells. */
+  const FilterPipeline& cell_var_offsets_filters() const;
+
+  /** Return the filter pipeline used for validity cells. */
+  const FilterPipeline& cell_validity_filters() const;
 
   /**
    * Checks the correctness of the array schema.
@@ -164,23 +150,14 @@ class ArraySchema {
    */
   Status check_attributes(const std::vector<std::string>& attributes) const;
 
-  /** Return the filter pipeline for the given attribute. */
-  const FilterPipeline* filters(const std::string& attribute) const;
+  /**
+   * Return the filter pipeline for the given attribute/dimension (can be
+   * TILEDB_COORDS).
+   */
+  const FilterPipeline& filters(const std::string& name) const;
 
-  /** Return a pointer to the pipeline used for coordinates. */
-  const FilterPipeline* coords_filters() const;
-
-  /** Returns the compressor of the coordinates. */
-  Compressor coords_compression() const;
-
-  /** Returns the compression level of the coordinates. */
-  int coords_compression_level() const;
-
-  /** Returns the coordinates size. */
-  uint64_t coords_size() const;
-
-  /** Returns the type of the coordinates. */
-  Datatype coords_type() const;
+  /** Return the pipeline used for coordinates. */
+  const FilterPipeline& coords_filters() const;
 
   /** True if the array is dense. */
   bool dense() const;
@@ -188,6 +165,18 @@ class ArraySchema {
   /** Returns the i-th dimension. */
   const Dimension* dimension(unsigned int i) const;
 
+  /**
+   * Returns a constant pointer to the selected dimension (nullptr if it
+   * does not exist).
+   */
+  const Dimension* dimension(const std::string& name) const;
+
+  /** Returns the dimension names. */
+  std::vector<std::string> dim_names() const;
+
+  /** Returns the dimension types. */
+  std::vector<Datatype> dim_types() const;
+
   /** Returns the number of dimensions. */
   unsigned int dim_num() const;
 
@@ -204,6 +193,18 @@ class ArraySchema {
    */
   Status has_attribute(const std::string& name, bool* has_attr) const;
 
+  /** Returns true if the input name is an attribute. */
+  bool is_attr(const std::string& name) const;
+
+  /** Returns true if the input name is a dimension. */
+  bool is_dim(const std::string& name) const;
+
+  /** Returns true if the input name is a dimension, attribute or coords. */
+  bool is_field(const std::string& name) const;
+
+  /** Returns true if the input name is nullable. */
+  bool is_nullable(const std::string& name) const;
+
   /**
    * Serializes the array schema object into a buffer.
    *
@@ -215,14 +216,17 @@ class ArraySchema {
   /** Returns the tile order. */
   Layout tile_order() const;
 
-  /** Returns the type of the i-th attribute. */
-  Datatype type(unsigned int i) const;
+  /**
+   * Returns the type of the input attribute/dimension (could also be
+   * TILEDB_COORDS).
+   */
+  Datatype type(const std::string& name) const;
 
-  /** Returns the type of the input attribute (could be coordinates). */
-  Datatype type(const std::string& attribute) const;
-
-  /** Returns *true* if the indicated attribute has variable-sized values. */
-  bool var_size(const std::string& attribute) const;
+  /**
+   * Returns *true* if the input attribute/dimension has variable-sized
+   * values.
+   */
+  bool var_size(const std::string& name) const;
 
   /**
    * Adds an attribute, copying the input.
@@ -237,6 +241,14 @@ class ArraySchema {
    */
   Status add_attribute(const Attribute* attr, bool check_special = true);
 
+  /**
+   * Drops an attribute.
+   *
+   * @param attr_name The name of the attribute to be removed.
+   * @return Status
+   */
+  Status drop_attribute(const std::string& attr_name);
+
   /**
    * It assigns values to the members of the object from the input buffer.
    *
@@ -256,12 +268,21 @@ class ArraySchema {
    */
   Status init();
 
+  /**
+   * Sets whether the array allows coordinate duplicates.
+   * It errors out if set to `1` for dense arrays.
+   */
+  Status set_allows_dups(bool allows_dups);
+
   /** Sets an array URI. */
   void set_array_uri(const URI& array_uri);
 
   /** Sets the filter pipeline for the variable cell offsets. */
   Status set_cell_var_offsets_filter_pipeline(const FilterPipeline* pipeline);
 
+  /** Sets the filter pipeline for the validity cell offsets. */
+  Status set_cell_validity_filter_pipeline(const FilterPipeline* pipeline);
+
   /** Sets the filter pipeline for the coordinates. */
   Status set_coords_filter_pipeline(const FilterPipeline* pipeline);
 
@@ -269,7 +290,7 @@ class ArraySchema {
   void set_capacity(uint64_t capacity);
 
   /** Sets the cell order. */
-  void set_cell_order(Layout cell_order);
+  Status set_cell_order(Layout cell_order);
 
   /**
    * Sets the domain. The function returns an error if the array has been
@@ -278,16 +299,56 @@ class ArraySchema {
   Status set_domain(Domain* domain);
 
   /** Sets the tile order. */
-  void set_tile_order(Layout tile_order);
+  Status set_tile_order(Layout tile_order);
+
+  /** Set version of schema, only used for serialization */
+  void set_version(uint32_t version);
+
+  /** Returns the version to write in. */
+  uint32_t write_version() const;
 
   /** Returns the array schema version. */
   uint32_t version() const;
 
+  /** Set a timestamp range for the array schema */
+  Status set_timestamp_range(
+      const std::pair<uint64_t, uint64_t>& timestamp_range);
+
+  /** Returns the timestamp range. */
+  std::pair<uint64_t, uint64_t> timestamp_range() const;
+
+  /** Returns the the first timestamp. */
+  uint64_t timestamp_start() const;
+
+  /** Returns the array schema uri. */
+  URI uri();
+
+  /** Set schema URI, along with parsing out timestamp ranges and name. */
+  void set_uri(const URI& uri);
+
+  /** Get schema URI with return status. */
+  Status get_uri(URI* uri);
+
+  /** Returns the schema name. If it is not set, will build it. */
+  std::string name();
+
+  /** Returns the schema name. If it is not set, will returns error status. */
+  Status get_name(std::string* name) const;
+
+  /** Generates a new array schema URI. */
+  Status generate_uri();
+
  private:
   /* ********************************* */
   /*         PRIVATE ATTRIBUTES        */
   /* ********************************* */
 
+  /**
+   * True if the array allows coordinate duplicates. Applicable only
+   * to sparse arrays.
+   */
+  bool allows_dups_;
+
   /** An array name attached to the schema object. */
   URI array_uri_;
 
@@ -311,17 +372,17 @@ class ArraySchema {
    */
   Layout cell_order_;
 
-  /** Stores the size of every attribute (plus coordinates). */
-  std::unordered_map<std::string, uint64_t> cell_sizes_;
-
   /** The filter pipeline run on offset tiles for var-length attributes. */
   FilterPipeline cell_var_offsets_filters_;
 
+  /** The filter pipeline run on validity tiles for nullable attributes. */
+  FilterPipeline cell_validity_filters_;
+
   /** The filter pipeline run on coordinate tiles. */
   FilterPipeline coords_filters_;
 
-  /** The size (in bytes) of the coordinates. */
-  uint64_t coords_size_;
+  /** It maps each dimension name to the corresponding dimension object. */
+  std::unordered_map<std::string, const Dimension*> dim_map_;
 
   /** The array domain. */
   Domain* domain_;
@@ -336,6 +397,23 @@ class ArraySchema {
   /** The format version of this array schema. */
   uint32_t version_;
 
+  /** Mutex for thread-safety. */
+  mutable std::mutex mtx_;
+
+  /**
+   * The timestamp the array schema was written.
+   * This is used to determine the array schema file name.
+   * The two timestamps are identical.
+   * It is stored as a pair to keep the usage consistent with metadata
+   */
+  std::pair<uint64_t, uint64_t> timestamp_range_;
+
+  /** The URI of the array schema file. */
+  URI uri_;
+
+  /** The file name of array schema in the format of timestamp_timestamp_uuid */
+  std::string name_;
+
   /* ********************************* */
   /*           PRIVATE METHODS         */
   /* ********************************* */
@@ -347,19 +425,16 @@ class ArraySchema {
   bool check_attribute_dimension_names() const;
 
   /**
-   * Returns false if double delta compression is used with real attributes
-   * or coordinates and true otherwise.
+   * Returns error if double delta compression is used in the zipped
+   * coordinate filters and is inherited by a dimension.
    */
-  bool check_double_delta_compressor() const;
+  Status check_double_delta_compressor() const;
 
   /** Clears all members. Use with caution! */
   void clear();
-
-  /** Computes and returns the size of an attribute (or coordinates). */
-  uint64_t compute_cell_size(const std::string& attribute) const;
 };
 
 }  // namespace sm
 }  // namespace tiledb
 
-#endif  // TILEDB_ARRAY_METADATA_H
+#endif  // TILEDB_ARRAY_SCHEMA_H
