@@ -5,7 +5,7 @@
  *
  * The MIT License
  *
- * @copyright Copyright (c) 2017-2019 TileDB, Inc.
+ * @copyright Copyright (c) 2017-2021 TileDB, Inc.
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
  * of this software and associated documentation files (the "Software"), to deal
@@ -33,17 +33,33 @@
 #ifndef TILEDB_VFS_H
 #define TILEDB_VFS_H
 
+#include <functional>
+#include <list>
+#include <set>
+#include <string>
+#include <vector>
+
+#include "tiledb/common/macros.h"
+#include "tiledb/common/status.h"
+#include "tiledb/common/thread_pool.h"
 #include "tiledb/sm/buffer/buffer.h"
+#include "tiledb/sm/cache/lru_cache.h"
 #include "tiledb/sm/config/config.h"
-#include "tiledb/sm/enums/filesystem.h"
-#include "tiledb/sm/enums/vfs_mode.h"
 #include "tiledb/sm/filesystem/filelock.h"
-#include "tiledb/sm/filesystem/posix.h"
-#include "tiledb/sm/filesystem/win.h"
+#include "tiledb/sm/filesystem/mem_filesystem.h"
 #include "tiledb/sm/misc/cancelable_tasks.h"
-#include "tiledb/sm/misc/status.h"
-#include "tiledb/sm/misc/thread_pool.h"
 #include "tiledb/sm/misc/uri.h"
+#include "tiledb/sm/stats/stats.h"
+
+#ifdef _WIN32
+#include "tiledb/sm/filesystem/win.h"
+#else
+#include "tiledb/sm/filesystem/posix.h"
+#endif
+
+#ifdef HAVE_GCS
+#include "tiledb/sm/filesystem/gcs.h"
+#endif
 
 #ifdef HAVE_S3
 #include "tiledb/sm/filesystem/s3.h"
@@ -53,14 +69,18 @@
 #include "tiledb/sm/filesystem/hdfs_filesystem.h"
 #endif
 
-#include <functional>
-#include <set>
-#include <string>
-#include <vector>
+#ifdef HAVE_AZURE
+#include "tiledb/sm/filesystem/azure.h"
+#endif
+
+using namespace tiledb::common;
 
 namespace tiledb {
 namespace sm {
 
+enum class Filesystem : uint8_t;
+enum class VFSMode : uint8_t;
+
 /**
  * This class implements a virtual filesystem that directs filesystem-related
  * function execution to the appropriate backend based on the input URI.
@@ -77,6 +97,9 @@ class VFS {
   /** Destructor. */
   ~VFS() = default;
 
+  DISABLE_COPY_AND_COPY_ASSIGN(VFS);
+  DISABLE_MOVE_AND_MOVE_ASSIGN(VFS);
+
   /* ********************************* */
   /*               API                 */
   /* ********************************* */
@@ -122,7 +145,7 @@ class VFS {
   Status cancel_all_tasks();
 
   /**
-   * Creates an object-store bucket.
+   * Creates an object store bucket.
    *
    * @param uri The name of the bucket to be created.
    * @return Status
@@ -143,7 +166,7 @@ class VFS {
   Status dir_size(const URI& dir_name, uint64_t* dir_size) const;
 
   /**
-   * Deletes an object-store bucket.
+   * Deletes an object store bucket.
    *
    * @param uri The name of the bucket to be deleted.
    * @return Status
@@ -151,7 +174,7 @@ class VFS {
   Status remove_bucket(const URI& uri) const;
 
   /**
-   * Deletes the contents of an object-store bucket.
+   * Deletes the contents of an object store bucket.
    *
    * @param uri The name of the bucket to be emptied.
    * @return Status
@@ -227,9 +250,9 @@ class VFS {
   Status is_file(const URI& uri, bool* is_file) const;
 
   /**
-   * Checks if an object-store bucket exists.
+   * Checks if an object store bucket exists.
    *
-   * @param uri The name of the S3 bucket.
+   * @param uri The name of the object store bucket.
    * @return is_bucket Set to `true` if the bucket exists and `false` otherwise.
    * @return Status
    */
@@ -238,7 +261,7 @@ class VFS {
   /**
    * Checks if an object-store bucket is empty.
    *
-   * @param uri The name of the S3 bucket.
+   * @param uri The name of the object store bucket.
    * @param is_empty Set to `true` if the bucket is empty and `false` otherwise.
    */
   Status is_empty_bucket(const URI& uri, bool* is_empty) const;
@@ -246,10 +269,16 @@ class VFS {
   /**
    * Initializes the virtual filesystem with the given configuration.
    *
+   * @param parent_stats The parent stats to inherit from.
    * @param config Configuration parameters
    * @return Status
    */
-  Status init(const Config* ctx_config, const Config* vfs_config);
+  Status init(
+      stats::Stats* parent_stats,
+      ThreadPool* compute_tp,
+      ThreadPool* io_tp,
+      const Config* ctx_config,
+      const Config* vfs_config);
 
   /**
    * Terminates the virtual system. Must only be called if init() returned
@@ -287,6 +316,24 @@ class VFS {
    */
   Status move_dir(const URI& old_uri, const URI& new_uri);
 
+  /**
+   * Copies a file.
+   *
+   * @param old_uri The old URI.
+   * @param new_uri The new URI.
+   * @return Status
+   */
+  Status copy_file(const URI& old_uri, const URI& new_uri);
+
+  /**
+   * Copies directory.
+   *
+   * @param old_uri The old URI.
+   * @param new_uri The new URI.
+   * @return Status
+   */
+  Status copy_dir(const URI& old_uri, const URI& new_uri);
+
   /**
    * Reads from a file.
    *
@@ -294,9 +341,15 @@ class VFS {
    * @param offset The offset where the read begins.
    * @param buffer The buffer to read into.
    * @param nbytes Number of bytes to read.
+   * @param use_read_ahead Whether to use the read-ahead cache.
    * @return Status
    */
-  Status read(const URI& uri, uint64_t offset, void* buffer, uint64_t nbytes);
+  Status read(
+      const URI& uri,
+      uint64_t offset,
+      void* buffer,
+      uint64_t nbytes,
+      bool use_read_ahead = true);
 
   /**
    * Reads multiple regions from a file.
@@ -304,15 +357,17 @@ class VFS {
    * @param uri The URI of the file.
    * @param regions The list of regions to read. Each region is a tuple
    *    `(file_offset, dest_buffer, nbytes)`.
-   * @param thread_pool Thread pool to enqueue async read tasks to.
+   * @param thread_pool Thread pool to execute async read tasks to.
    * @param tasks Vector to which new async read tasks are pushed.
+   * @param use_read_ahead Whether to use the read-ahead cache.
    * @return Status
    */
   Status read_all(
       const URI& uri,
       const std::vector<std::tuple<uint64_t, void*, uint64_t>>& regions,
       ThreadPool* thread_pool,
-      std::vector<std::future<Status>>* tasks);
+      std::vector<ThreadPool::Task>* tasks,
+      bool use_read_ahead = true);
 
   /** Checks if a given filesystem is supported. */
   bool supports_fs(Filesystem fs) const;
@@ -396,10 +451,171 @@ class VFS {
     std::vector<std::tuple<uint64_t, void*, uint64_t>> regions;
   };
 
+  /**
+   * Represents a sub-range of data within a URI file at a
+   * specific file offset.
+   */
+  struct ReadAheadBuffer {
+    /* ********************************* */
+    /*            CONSTRUCTORS           */
+    /* ********************************* */
+
+    /** Value Constructor. */
+    ReadAheadBuffer(const uint64_t offset, Buffer&& buffer)
+        : offset_(offset)
+        , buffer_(std::move(buffer)) {
+    }
+
+    /** Move Constructor. */
+    ReadAheadBuffer(ReadAheadBuffer&& other)
+        : offset_(other.offset_)
+        , buffer_(std::move(other.buffer_)) {
+    }
+
+    /* ********************************* */
+    /*             OPERATORS             */
+    /* ********************************* */
+
+    /** Move-Assign Operator. */
+    ReadAheadBuffer& operator=(ReadAheadBuffer&& other) {
+      offset_ = other.offset_;
+      buffer_ = std::move(other.buffer_);
+      return *this;
+    }
+
+    DISABLE_COPY_AND_COPY_ASSIGN(ReadAheadBuffer);
+
+    /* ********************************* */
+    /*             ATTRIBUTES            */
+    /* ********************************* */
+
+    /** The offset within the associated URI. */
+    uint64_t offset_;
+
+    /** The buffered data at `offset`. */
+    Buffer buffer_;
+  };
+
+  /**
+   * An LRU cache of `ReadAheadBuffer` objects keyed by a URI string.
+   */
+  class ReadAheadCache : public LRUCache<std::string, ReadAheadBuffer> {
+   public:
+    /* ********************************* */
+    /*     CONSTRUCTORS & DESTRUCTORS    */
+    /* ********************************* */
+
+    /** Constructor. */
+    ReadAheadCache(const uint64_t max_cached_buffers)
+        : LRUCache(max_cached_buffers) {
+    }
+
+    /** Destructor. */
+    virtual ~ReadAheadCache() = default;
+
+    /* ********************************* */
+    /*                API                */
+    /* ********************************* */
+
+    /**
+     * Attempts to read a buffer from the cache.
+     *
+     * @param uri The URI associated with the buffer to cache.
+     * @param offset The offset that buffer starts at within the URI.
+     * @param buffer The buffer to cache.
+     * @param nbytes The number of bytes within the buffer.
+     * @param success True if `buffer` was read from the cache.
+     * @return Status
+     */
+    Status read(
+        const URI& uri,
+        const uint64_t offset,
+        void* const buffer,
+        const uint64_t nbytes,
+        bool* const success) {
+      assert(success);
+      *success = false;
+
+      // Store the URI's string representation.
+      const std::string uri_str = uri.to_string();
+
+      // Protect access to the derived LRUCache routines.
+      std::lock_guard<std::mutex> lg(lru_mtx_);
+
+      // Check that a cached buffer exists for `uri`.
+      if (!has_item(uri_str))
+        return Status::Ok();
+
+      // Store a reference to the cached buffer.
+      const ReadAheadBuffer* const ra_buffer = get_item(uri_str);
+
+      // Check that the read offset is not below the offset of
+      // the cached buffer.
+      if (offset < ra_buffer->offset_)
+        return Status::Ok();
+
+      // Calculate the offset within the cached buffer that corresponds
+      // to the requested read offset.
+      const uint64_t offset_in_buffer = offset - ra_buffer->offset_;
+
+      // Check that both the start and end positions of the requested
+      // read range reside within the cached buffer.
+      if (offset_in_buffer + nbytes > ra_buffer->buffer_.size())
+        return Status::Ok();
+
+      // Copy the subrange of the cached buffer that satisfies the caller's
+      // read request back into their output `buffer`.
+      std::memcpy(
+          buffer,
+          static_cast<uint8_t*>(ra_buffer->buffer_.data()) + offset_in_buffer,
+          nbytes);
+
+      // Touch the item to make it the most recently used item.
+      touch_item(uri_str);
+
+      *success = true;
+      return Status::Ok();
+    }
+
+    /**
+     * Writes a cached buffer for the given uri.
+     *
+     * @param uri The URI associated with the buffer to cache.
+     * @param offset The offset that buffer starts at within the URI.
+     * @param buffer The buffer to cache.
+     * @return Status
+     */
+    Status insert(const URI& uri, const uint64_t offset, Buffer&& buffer) {
+      // Protect access to the derived LRUCache routines.
+      std::lock_guard<std::mutex> lg(lru_mtx_);
+
+      const uint64_t size = buffer.size();
+      ReadAheadBuffer ra_buffer(offset, std::move(buffer));
+      return LRUCache<std::string, ReadAheadBuffer>::insert(
+          uri.to_string(), std::move(ra_buffer), size);
+    }
+
+   private:
+    /* ********************************* */
+    /*         PRIVATE ATTRIBUTES        */
+    /* ********************************* */
+
+    // Protects LRUCache routines.
+    std::mutex lru_mtx_;
+  };
+
   /* ********************************* */
   /*         PRIVATE ATTRIBUTES        */
   /* ********************************* */
 
+#ifdef HAVE_AZURE
+  Azure azure_;
+#endif
+
+#ifdef HAVE_GCS
+  GCS gcs_;
+#endif
+
 #ifdef HAVE_S3
   S3 s3_;
 #endif
@@ -411,24 +627,43 @@ class VFS {
 #endif
 
 #ifdef HAVE_HDFS
-  std::unique_ptr<hdfs::HDFS> hdfs_;
+  tdb_unique_ptr<hdfs::HDFS> hdfs_;
 #endif
 
+  /** The class stats. */
+  stats::Stats* stats_;
+
+  /** The in-memory filesystem which is always supported */
+  MemFilesystem memfs_;
+
   /** Config. */
   Config config_;
 
   /** `true` if the VFS object has been initialized. */
   bool init_;
 
+  /** The byte size to read-ahead for each read. */
+  uint64_t read_ahead_size_;
+
   /** The set with the supported filesystems. */
   std::set<Filesystem> supported_fs_;
 
-  /** Thread pool for parallel I/O operations. */
-  ThreadPool thread_pool_;
+  /** Thread pool for compute-bound tasks. */
+  ThreadPool* compute_tp_;
+
+  /** Thread pool for io-bound tasks. */
+  ThreadPool* io_tp_;
 
   /** Wrapper for tracking and canceling certain tasks on 'thread_pool' */
   CancelableTasks cancelable_tasks_;
 
+  /** The read-ahead cache. */
+  tdb_unique_ptr<ReadAheadCache> read_ahead_cache_;
+
+  /* ********************************* */
+  /*          PRIVATE METHODS          */
+  /* ********************************* */
+
   /**
    * Groups the given vector of regions to be read into a possibly smaller
    * vector of batched reads.
@@ -449,10 +684,35 @@ class VFS {
    * @param offset The offset where the read begins.
    * @param buffer The buffer to read into.
    * @param nbytes Number of bytes to read.
+   * @param use_read_ahead Whether to use the read-ahead cache.
    * @return Status
    */
   Status read_impl(
-      const URI& uri, uint64_t offset, void* buffer, uint64_t nbytes);
+      const URI& uri,
+      uint64_t offset,
+      void* buffer,
+      uint64_t nbytes,
+      bool use_read_ahead);
+
+  /**
+   * Executes a read, using the read-ahead cache as necessary.
+   *
+   * @param read_fn The read routine to execute.
+   * @param uri The URI of the file.
+   * @param offset The offset where the read begins.
+   * @param buffer The buffer to read into.
+   * @param nbytes Number of bytes to read.
+   * @param use_read_ahead Whether to use the read-ahead cache.
+   * @return Status
+   */
+  Status read_ahead_impl(
+      const std::function<Status(
+          const URI&, off_t, void*, uint64_t, uint64_t, uint64_t*)>& read_fn,
+      const URI& uri,
+      const uint64_t offset,
+      void* const buffer,
+      const uint64_t nbytes,
+      const bool use_read_ahead);
 
   /**
    * Decrement the lock count of the given URI.
