@@ -5,7 +5,7 @@
  *
  * The MIT License
  *
- * @copyright Copyright (c) 2017-2019 TileDB, Inc.
+ * @copyright Copyright (c) 2017-2021 TileDB, Inc.
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
  * of this software and associated documentation files (the "Software"), to deal
@@ -31,22 +31,28 @@
  */
 
 #include "tiledb/sm/filter/filter_pipeline.h"
+#include "tiledb/common/heap_memory.h"
+#include "tiledb/common/logger.h"
+#include "tiledb/sm/crypto/encryption_key.h"
+#include "tiledb/sm/enums/encryption_type.h"
+#include "tiledb/sm/enums/filter_type.h"
 #include "tiledb/sm/filter/compression_filter.h"
 #include "tiledb/sm/filter/encryption_aes256gcm_filter.h"
 #include "tiledb/sm/filter/filter.h"
 #include "tiledb/sm/filter/filter_storage.h"
 #include "tiledb/sm/filter/noop_filter.h"
-#include "tiledb/sm/misc/logger.h"
 #include "tiledb/sm/misc/parallel_functions.h"
-#include "tiledb/sm/misc/stats.h"
+#include "tiledb/sm/stats/global_stats.h"
 #include "tiledb/sm/tile/tile.h"
 
+using namespace tiledb::common;
+
 namespace tiledb {
 namespace sm {
 
-FilterPipeline::FilterPipeline() {
-  current_tile_ = nullptr;
-  max_chunk_size_ = constants::max_tile_chunk_size;
+FilterPipeline::FilterPipeline()
+    : current_tile_(nullptr)
+    , max_chunk_size_(constants::max_tile_chunk_size) {
 }
 
 FilterPipeline::FilterPipeline(const FilterPipeline& other) {
@@ -75,7 +81,7 @@ FilterPipeline& FilterPipeline::operator=(FilterPipeline&& other) {
 }
 
 Status FilterPipeline::add_filter(const Filter& filter) {
-  std::unique_ptr<Filter> copy(filter.clone());
+  tdb_unique_ptr<Filter> copy(filter.clone());
   copy->set_pipeline(this);
   filters_.push_back(std::move(copy));
   return Status::Ok();
@@ -85,58 +91,32 @@ void FilterPipeline::clear() {
   filters_.clear();
 }
 
-Status FilterPipeline::compute_tile_chunks(
-    Tile* tile, std::vector<std::pair<void*, uint32_t>>* chunks) const {
-  // For coordinate tiles, we treat each dimension separately (chunks won't
-  // cross dimension boundaries, since the coordinates have been split).
-  // Attribute tiles are treated as a whole.
-  auto dim_num = tile->stores_coords() ? tile->dim_num() : 1;
-  auto dim_tile_size = tile->size() / dim_num;
-  auto dim_cell_size = tile->cell_size() / dim_num;
-
-  // Compute a chunk size as a multiple of the cell size, ensuring that the
-  // chunk contains always at least 1 cell.
-  uint64_t chunk_size = std::min((uint64_t)max_chunk_size_, dim_tile_size);
-  chunk_size = chunk_size / dim_cell_size * dim_cell_size;
-  chunk_size = std::max(chunk_size, dim_cell_size);
-  if (chunk_size > std::numeric_limits<uint32_t>::max())
-    return LOG_STATUS(
-        Status::FilterError("Filter error; chunk size exceeds uint32_t"));
-
-  // Compute number of chunks
-  auto chunk_num = dim_tile_size / (chunk_size) +
-                   uint64_t(bool(dim_tile_size % (chunk_size)));
-
-  // Compute the chunks
-  tile->reset_offset();
-  for (uint64_t i = 0; i < dim_num; i++) {
-    for (uint64_t j = 0; j < chunk_num; j++) {
-      // Compute the actual size of the chunk (may be smaller at the end of the
-      // tile if the chunk size doesn't evenly divide).
-      auto size = static_cast<uint32_t>(
-          std::min(chunk_size, dim_tile_size - j * chunk_size));
-      chunks->emplace_back(tile->cur_data(), size);
-      tile->advance_offset(size);
-    }
-  }
-
-  return Status::Ok();
-}
-
 const Tile* FilterPipeline::current_tile() const {
   return current_tile_;
 }
 
 Status FilterPipeline::filter_chunks_forward(
-    const std::vector<std::pair<void*, uint32_t>>& chunks,
-    Buffer* output) const {
+    const Buffer& input,
+    uint32_t chunk_size,
+    Buffer* const output,
+    ThreadPool* const compute_tp) const {
+  assert(output);
+
+  uint64_t nchunks = input.size() / chunk_size;
+  uint64_t last_buffer_size = input.size() % chunk_size;
+  if (last_buffer_size != 0) {
+    nchunks++;
+  } else {
+    last_buffer_size = chunk_size;
+  }
+
   // Vector storing the input and output of the final pipeline stage for each
   // chunk.
   std::vector<std::pair<FilterBufferPair, FilterBufferPair>> final_stage_io(
-      chunks.size());
+      nchunks);
 
   // Run each chunk through the entire pipeline.
-  auto statuses = parallel_for(0, chunks.size(), [&](uint64_t i) {
+  auto status = parallel_for(compute_tp, 0, nchunks, [&](uint64_t i) {
     // TODO(ttd): can we instead allocate one FilterStorage per thread?
     // or make it threadsafe?
     FilterStorage storage;
@@ -144,8 +124,10 @@ Status FilterPipeline::filter_chunks_forward(
     FilterBuffer input_metadata(&storage), output_metadata(&storage);
 
     // First filter's input is the original chunk.
-    const auto& chunk_input = chunks[i];
-    RETURN_NOT_OK(input_data.init(chunk_input.first, chunk_input.second));
+    void* chunk_buffer = static_cast<char*>(input.data()) + i * chunk_size;
+    uint32_t chunk_buffer_size =
+        i == nchunks - 1 ? last_buffer_size : chunk_size;
+    RETURN_NOT_OK(input_data.init(chunk_buffer, chunk_buffer_size));
 
     // Apply the filters sequentially.
     for (auto it = filters_.begin(), ite = filters_.end(); it != ite; ++it) {
@@ -170,11 +152,12 @@ Status FilterPipeline::filter_chunks_forward(
       // Next input (input_buffers) now stores this output (output_buffers).
     }
 
-    // Save the finished chunk (last stage's output). This is safe to do because
-    // when the local FilterStorage goes out of scope, it will not free the
-    // buffers saved here as their shared_ptr counters will not be zero.
-    // However, as the output may have been a view on the input, we do need to
-    // save both here to prevent the input buffer from being freed.
+    // Save the finished chunk (last stage's output). This is safe to do
+    // because when the local FilterStorage goes out of scope, it will not
+    // free the buffers saved here as their tdb_shared_ptr counters will not
+    // be zero. However, as the output may have been a view on the input, we
+    // do need to save both here to prevent the input buffer from being
+    // freed.
     auto& io = final_stage_io[i];
     auto& io_input = io.first;
     auto& io_output = io.second;
@@ -182,18 +165,14 @@ Status FilterPipeline::filter_chunks_forward(
     io_input.second.swap(input_data);
     io_output.first.swap(output_metadata);
     io_output.second.swap(output_data);
-
     return Status::Ok();
   });
 
-  // Check statuses
-  for (auto st : statuses)
-    RETURN_NOT_OK(st);
+  RETURN_NOT_OK(status);
 
-  // Compute the destination offset of each processed chunk in the final output
-  // buffer.
-  uint64_t offset = output->offset();
   uint64_t total_processed_size = 0;
+  std::vector<uint32_t> var_chunk_sizes(final_stage_io.size());
+  uint64_t offset = sizeof(uint64_t);
   std::vector<uint64_t> offsets(final_stage_io.size());
   for (uint64_t i = 0; i < final_stage_io.size(); i++) {
     auto& final_stage_output_metadata = final_stage_io[i].first.first;
@@ -207,21 +186,30 @@ Status FilterPipeline::filter_chunks_forward(
           "Filter error; filtered chunk size exceeds uint32_t"));
 
     // Leave space for the chunk sizes and the data itself.
-    auto space_required = 3 * sizeof(uint32_t) +
-                          final_stage_output_data.size() +
-                          final_stage_output_metadata.size();
+    const uint32_t space_required = 3 * sizeof(uint32_t) +
+                                    final_stage_output_data.size() +
+                                    final_stage_output_metadata.size();
+
+    total_processed_size += space_required;
+    var_chunk_sizes[i] = space_required;
     offsets[i] = offset;
     offset += space_required;
-    total_processed_size += space_required;
   }
 
+  // Allocate enough space in 'output' to store the leading uint64_t
+  // prefix containing the number of chunks and the 'total_processed_size'.
+  RETURN_NOT_OK(output->realloc(sizeof(uint64_t) + total_processed_size));
+
+  // Write the leading prefix that contains the number of chunks.
+  RETURN_NOT_OK(output->write(&nchunks, sizeof(uint64_t)));
+
   // Concatenate all processed chunks into the final output buffer.
-  RETURN_NOT_OK(output->realloc(output->size() + total_processed_size));
-  statuses = parallel_for(0, final_stage_io.size(), [&](uint64_t i) {
+  status = parallel_for(compute_tp, 0, final_stage_io.size(), [&](uint64_t i) {
     auto& final_stage_output_metadata = final_stage_io[i].first.first;
     auto& final_stage_output_data = final_stage_io[i].first.second;
     auto filtered_size = (uint32_t)final_stage_output_data.size();
-    auto orig_chunk_size = chunks[i].second;
+    uint32_t orig_chunk_size =
+        i == final_stage_io.size() - 1 ? last_buffer_size : chunk_size;
     auto metadata_size = (uint32_t)final_stage_output_metadata.size();
     void* dest = output->data(offsets[i]);
     uint64_t dest_offset = 0;
@@ -244,9 +232,7 @@ Status FilterPipeline::filter_chunks_forward(
     return Status::Ok();
   });
 
-  // Check statuses
-  for (auto st : statuses)
-    RETURN_NOT_OK(st);
+  RETURN_NOT_OK(status);
 
   // Ensure the final size is set to the concatenated size.
   output->advance_offset(total_processed_size);
@@ -256,24 +242,33 @@ Status FilterPipeline::filter_chunks_forward(
 }
 
 Status FilterPipeline::filter_chunks_reverse(
-    const std::vector<std::tuple<void*, uint32_t, uint32_t, uint32_t>>& chunks,
-    Buffer* output) const {
-  // Precompute the offsets for the final chunks in the shared output buffer.
-  std::vector<uint64_t> chunk_dest_offsets(chunks.size());
-  uint64_t chunk_dest_offset = 0;
-  for (uint64_t i = 0; i < chunks.size(); i++) {
-    chunk_dest_offsets[i] = chunk_dest_offset;
-    chunk_dest_offset += std::get<2>(chunks[i]);
+    const std::vector<std::tuple<void*, uint32_t, uint32_t, uint32_t>>& input,
+    Buffer* const output,
+    ThreadPool* const compute_tp,
+    const Config& config) const {
+  if (input.empty()) {
+    return Status::Ok();
   }
 
+  // Precompute the sizes of the final output chunks.
+  uint64_t total_size = 0;
+  std::vector<uint64_t> chunk_offsets(input.size());
+  for (size_t i = 0; i < input.size(); i++) {
+    chunk_offsets[i] = total_size;
+    total_size += std::get<2>(input[i]);
+  }
+
+  RETURN_NOT_OK(output->realloc(total_size));
+
   // Run each chunk through the entire pipeline.
-  auto statuses = parallel_for(0, chunks.size(), [&](uint64_t i) {
-    const auto& chunk_input = chunks[i];
-    uint32_t filtered_chunk_len = std::get<1>(chunk_input);
-    uint32_t orig_chunk_len = std::get<2>(chunk_input);
-    uint32_t metadata_len = std::get<3>(chunk_input);
-    void* metadata = std::get<0>(chunk_input);
-    void* chunk_data = (char*)metadata + metadata_len;
+  auto status = parallel_for(compute_tp, 0, input.size(), [&](uint64_t i) {
+    const auto& chunk_input = input[i];
+
+    const uint32_t filtered_chunk_len = std::get<1>(chunk_input);
+    const uint32_t orig_chunk_len = std::get<2>(chunk_input);
+    const uint32_t metadata_len = std::get<3>(chunk_input);
+    void* const metadata = std::get<0>(chunk_input);
+    void* const chunk_data = (char*)metadata + metadata_len;
 
     // TODO(ttd): can we instead allocate one FilterStorage per thread?
     // or make it threadsafe?
@@ -287,7 +282,9 @@ Status FilterPipeline::filter_chunks_reverse(
 
     // If the pipeline is empty, just copy input to output.
     if (filters_.empty()) {
-      RETURN_NOT_OK(input_data.copy_to(output->data(chunk_dest_offsets[i])));
+      void* output_chunk_buffer =
+          static_cast<char*>(output->data()) + chunk_offsets[i];
+      RETURN_NOT_OK(input_data.copy_to(output_chunk_buffer));
       return Status::Ok();
     }
 
@@ -308,12 +305,18 @@ Status FilterPipeline::filter_chunks_reverse(
       // Final filter: output directly into the shared output buffer.
       bool last_filter = filter_idx == 0;
       if (last_filter) {
-        void* dest = output->data(chunk_dest_offsets[i]);
-        RETURN_NOT_OK(output_data.set_fixed_allocation(dest, orig_chunk_len));
+        void* output_chunk_buffer =
+            static_cast<char*>(output->data()) + chunk_offsets[i];
+        RETURN_NOT_OK(output_data.set_fixed_allocation(
+            output_chunk_buffer, orig_chunk_len));
       }
 
       RETURN_NOT_OK(f->run_reverse(
-          &input_metadata, &input_data, &output_metadata, &output_data));
+          &input_metadata,
+          &input_data,
+          &output_metadata,
+          &output_data,
+          config));
 
       input_data.set_read_only(false);
       input_metadata.set_read_only(false);
@@ -328,13 +331,12 @@ Status FilterPipeline::filter_chunks_reverse(
     return Status::Ok();
   });
 
-  // Check statuses
-  for (auto st : statuses)
-    RETURN_NOT_OK(st);
+  RETURN_NOT_OK(status);
 
-  // Ensure the final size is set to the sum of unfiltered chunk sizes.
-  output->set_offset(chunk_dest_offset);
-  output->set_size(chunk_dest_offset);
+  // Since we did not use the 'write' interface above, the 'output' size
+  // will still be 0. We wrote the entire capacity of the output buffer,
+  // set it here.
+  output->set_size(total_size);
 
   return Status::Ok();
 }
@@ -350,96 +352,111 @@ uint32_t FilterPipeline::max_chunk_size() const {
   return max_chunk_size_;
 }
 
-Status FilterPipeline::run_forward(Tile* tile) const {
-  STATS_FUNC_IN(filter_pipeline_run_forward);
-
+Status FilterPipeline::run_forward(
+    stats::Stats* const writer_stats,
+    Tile* const tile,
+    ThreadPool* const compute_tp) const {
   current_tile_ = tile;
 
-  // Split the coords if the tile stores coordinates.
-  if (tile->stores_coords())
-    tile->split_coordinates();
+  writer_stats->add_counter("write_filtered_byte_num", tile->size());
 
-  // Compute the chunks.
-  std::vector<std::pair<void*, uint32_t>> chunks;
-  RETURN_NOT_OK(compute_tile_chunks(tile, &chunks));
-  uint64_t num_chunks = chunks.size();
-  if (num_chunks == 0)
-    return Status::FilterError("Filter error; tile has 0 chunks.");
+  uint32_t chunk_size = 0;
+  RETURN_NOT_OK(Tile::compute_chunk_size(
+      tile->size(), tile->dim_num(), tile->cell_size(), &chunk_size));
 
-  // Allocate a buffer to hold the end result (the concatentated, filtered
-  // chunks), and write the number of chunks.
-  Buffer filtered_tile;
-  filtered_tile.realloc(tile->buffer()->size());
-  RETURN_NOT_OK(filtered_tile.write(&num_chunks, sizeof(uint64_t)));
+  // Run the filters over all the chunks and store the result in
+  // 'filtered_buffer'.
+  RETURN_NOT_OK_ELSE(
+      filter_chunks_forward(
+          *tile->buffer(), chunk_size, tile->filtered_buffer(), compute_tp),
+      tile->filtered_buffer()->clear());
 
-  // Run the filters over all the chunks into the filtered_tile buffer.
-  RETURN_NOT_OK(filter_chunks_forward(chunks, &filtered_tile));
-
-  // Replace the tile's buffer with the filtered buffer.
-  RETURN_NOT_OK(tile->buffer()->swap(filtered_tile));
+  // The contents of 'buffer' have been filtered and stored
+  // in 'filtered_buffer'. We can safely free 'buffer'.
+  tile->buffer()->clear();
 
   return Status::Ok();
-
-  STATS_FUNC_OUT(filter_pipeline_run_forward);
 }
 
-Status FilterPipeline::run_reverse(Tile* tile) const {
-  STATS_FUNC_IN(filter_pipeline_run_reverse);
+Status FilterPipeline::run_reverse(
+    stats::Stats* const reader_stats,
+    Tile* const tile,
+    ThreadPool* const compute_tp,
+    const Config& config) const {
+  assert(tile->filtered());
 
-  auto tile_buff = tile->buffer();
-  if (tile_buff == nullptr)
+  return run_reverse_internal(reader_stats, tile, compute_tp, config);
+}
+
+Status FilterPipeline::run_reverse_internal(
+    stats::Stats* const reader_stats,
+    Tile* tile,
+    ThreadPool* const compute_tp,
+    const Config& config) const {
+  Buffer* const filtered_buffer = tile->filtered_buffer();
+  if (filtered_buffer == nullptr)
     return LOG_STATUS(
         Status::FilterError("Filter error; tile has null buffer."));
 
+  assert(tile->buffer());
+  assert(tile->buffer()->size() == 0);
+  if (tile->buffer()->size() > 0)
+    return LOG_STATUS(Status::FilterError(
+        "Filter error; tile has allocated uncompressed chunk buffers."));
+
   current_tile_ = tile;
 
   // First make a pass over the tile to get the chunk information.
-  tile_buff->reset_offset();
+  filtered_buffer->reset_offset();
   uint64_t num_chunks;
-  RETURN_NOT_OK(tile_buff->read(&num_chunks, sizeof(uint64_t)));
-  std::vector<std::tuple<void*, uint32_t, uint32_t, uint32_t>> chunks(
+  RETURN_NOT_OK(filtered_buffer->read(&num_chunks, sizeof(uint64_t)));
+  std::vector<std::tuple<void*, uint32_t, uint32_t, uint32_t>> filtered_chunks(
       num_chunks);
   uint64_t total_orig_size = 0;
   for (uint64_t i = 0; i < num_chunks; i++) {
     uint32_t filtered_chunk_size, orig_chunk_size, metadata_size;
-    RETURN_NOT_OK(tile_buff->read(&orig_chunk_size, sizeof(uint32_t)));
-    RETURN_NOT_OK(tile_buff->read(&filtered_chunk_size, sizeof(uint32_t)));
-    RETURN_NOT_OK(tile_buff->read(&metadata_size, sizeof(uint32_t)));
-    chunks[i] = std::make_tuple(
-        tile_buff->cur_data(),
+    RETURN_NOT_OK(filtered_buffer->read(&orig_chunk_size, sizeof(uint32_t)));
+    RETURN_NOT_OK(
+        filtered_buffer->read(&filtered_chunk_size, sizeof(uint32_t)));
+    RETURN_NOT_OK(filtered_buffer->read(&metadata_size, sizeof(uint32_t)));
+
+    total_orig_size += orig_chunk_size;
+
+    filtered_chunks[i] = std::make_tuple(
+        filtered_buffer->cur_data(),
         filtered_chunk_size,
         orig_chunk_size,
         metadata_size);
-    tile_buff->advance_offset(metadata_size + filtered_chunk_size);
-    total_orig_size += orig_chunk_size;
+    filtered_buffer->advance_offset(metadata_size + filtered_chunk_size);
   }
-  assert(tile_buff->offset() == tile_buff->size());
+  assert(filtered_buffer->offset() == filtered_buffer->size());
 
-  // Allocate a buffer to hold the end result (the assembled, unfiltered
-  // chunks).
-  Buffer unfiltered_tile;
-  RETURN_NOT_OK(unfiltered_tile.realloc(total_orig_size));
+  reader_stats->add_counter("read_unfiltered_byte_num", total_orig_size);
 
-  // Run the filters in reverse over all the chunks into the unfiltered_tile
-  // buffer.
-  RETURN_NOT_OK(filter_chunks_reverse(chunks, &unfiltered_tile));
+  const Status st = filter_chunks_reverse(
+      filtered_chunks, tile->buffer(), compute_tp, config);
+  if (!st.ok()) {
+    tile->buffer()->clear();
+    return st;
+  }
 
-  // Replace the tile's buffer with the unfiltered buffer.
-  RETURN_NOT_OK(tile->buffer()->swap(unfiltered_tile));
+  // Clear the filtered buffer now that we have reverse-filtered it into
+  // 'tile->buffer()'.
+  filtered_buffer->clear();
 
   // Zip the coords.
   if (tile->stores_coords()) {
     // Note that format version < 2 only split the coordinates when compression
     // was used. See https://github.com/TileDB-Inc/TileDB/issues/1053
+    // For format version > 4, a tile never stores coordinates
     bool using_compression = get_filter<CompressionFilter>() != nullptr;
-    if (tile->format_version() > 1 || using_compression) {
-      tile->zip_coordinates();
+    auto version = tile->format_version();
+    if (version > 1 || using_compression) {
+      RETURN_NOT_OK(tile->zip_coordinates());
     }
   }
 
   return Status::Ok();
-
-  STATS_FUNC_OUT(filter_pipeline_run_reverse);
 }
 
 // ===== FORMAT =====
@@ -458,7 +475,7 @@ Status FilterPipeline::serialize(Buffer* buff) const {
     // as a no-op filter instead.
     auto as_compression = dynamic_cast<CompressionFilter*>(f.get());
     if (as_compression != nullptr && f->type() == FilterType::FILTER_NONE) {
-      auto noop = std::unique_ptr<NoopFilter>(new NoopFilter);
+      auto noop = tdb_unique_ptr<NoopFilter>(new NoopFilter);
       RETURN_NOT_OK(noop->serialize(buff));
     } else {
       RETURN_NOT_OK(f->serialize(buff));
@@ -479,13 +496,23 @@ Status FilterPipeline::deserialize(ConstBuffer* buff) {
   for (uint32_t i = 0; i < num_filters; i++) {
     Filter* filter;
     RETURN_NOT_OK(Filter::deserialize(buff, &filter));
-    RETURN_NOT_OK_ELSE(add_filter(*filter), delete filter);
-    delete filter;
+    RETURN_NOT_OK_ELSE(add_filter(*filter), tdb_delete(filter));
+    tdb_delete(filter);
   }
 
   return Status::Ok();
 }
 
+void FilterPipeline::dump(FILE* out) const {
+  if (out == nullptr)
+    out = stdout;
+
+  for (const auto& filter : filters_) {
+    fprintf(out, "\n  > ");
+    filter->dump(out);
+  }
+}
+
 void FilterPipeline::set_max_chunk_size(uint32_t max_chunk_size) {
   max_chunk_size_ = max_chunk_size;
 }
@@ -494,6 +521,10 @@ unsigned FilterPipeline::size() const {
   return static_cast<unsigned>(filters_.size());
 }
 
+bool FilterPipeline::empty() const {
+  return filters_.empty();
+}
+
 void FilterPipeline::swap(FilterPipeline& other) {
   filters_.swap(other.filters_);
 
