@@ -5,7 +5,7 @@
  *
  * The MIT License
  *
- * @copyright Copyright (c) 2017-2019 TileDB, Inc.
+ * @copyright Copyright (c) 2017-2021 TileDB, Inc.
  * @copyright Copyright (c) 2016 MIT and Intel Corporation
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
@@ -32,8 +32,13 @@
  */
 
 #include "tiledb/sm/misc/utils.h"
-#include "tiledb/sm/misc/logger.h"
+#include "tiledb/common/logger.h"
+#include "tiledb/sm/enums/datatype.h"
+#include "tiledb/sm/enums/serialization_type.h"
+#include "tiledb/sm/misc/constants.h"
+#include "tiledb/sm/misc/uri.h"
 
+#include <algorithm>
 #include <iostream>
 #include <set>
 #include <sstream>
@@ -45,12 +50,11 @@
 #include <sys/time.h>
 #endif
 
-/* ****************************** */
-/*             MACROS             */
-/* ****************************** */
+#ifdef __linux__
+#include "tiledb/sm/filesystem/posix.h"
+#endif
 
-#define MIN(a, b) ((a) < (b) ? (a) : (b))
-#define MAX(a, b) ((a) > (b) ? (a) : (b))
+using namespace tiledb::common;
 
 namespace tiledb {
 namespace sm {
@@ -229,40 +233,83 @@ Status convert(const std::string& str, SerializationType* value) {
   return Status::Ok();
 }
 
-std::pair<uint64_t, uint64_t> get_timestamp_range(
-    uint32_t version, const std::string& fragment_name) {
-  std::pair<uint64_t, uint64_t> ret = {0, 0};
+Status get_timestamp_range(
+    const URI& uri, std::pair<uint64_t, uint64_t>* timestamp_range) {
+  // Initializations
+  auto name = uri.remove_trailing_slash().last_path_part();
+  *timestamp_range = {0, 0};
+
+  // Cut the suffix
+  auto pos = name.find_last_of('.');
+  name = (pos == std::string::npos) ? name : name.substr(0, pos);
+
+  // Get fragment version
+  uint32_t version = 0;
+  RETURN_NOT_OK(utils::parse::get_fragment_name_version(name, &version));
+
   if (version == 1) {  // This is equivalent to format version <=2
-    assert(fragment_name.find_last_of('_') != std::string::npos);
-    auto t_str = fragment_name.substr(fragment_name.find_last_of('_') + 1);
+    assert(name.find_last_of('_') != std::string::npos);
+    auto t_str = name.substr(name.find_last_of('_') + 1);
     sscanf(
         t_str.c_str(),
         (std::string("%") + std::string(PRId64)).c_str(),
-        (long long int*)&ret.first);
-    ret.second = ret.first;
+        (long long int*)&timestamp_range->first);
+    timestamp_range->second = timestamp_range->first;
   } else {
-    assert(fragment_name.find_last_of('_') != std::string::npos);
+    assert(name.find_last_of('_') != std::string::npos);
     sscanf(
-        fragment_name.c_str(),
+        name.c_str(),
         (std::string("__%") + std::string(PRId64) + "_%" + std::string(PRId64))
             .c_str(),
-        (long long int*)&ret.first,
-        (long long int*)&ret.second);
+        (long long int*)&timestamp_range->first,
+        (long long int*)&timestamp_range->second);
   }
 
-  return ret;
+  return Status::Ok();
 }
 
-Status get_fragment_name_version(
-    const std::string& fragment_name, uint32_t* version) {
-  auto t_str = fragment_name.substr(fragment_name.find_last_of('_') + 1);
+Status get_fragment_name_version(const std::string& name, uint32_t* version) {
+  // First check if it is version 3 or greater, which has 5 '_'
+  // characters in the name.
+  size_t n = std::count(name.begin(), name.end(), '_');
+  if (n == 5) {
+    // Fetch the fragment version from the fragment name. If the fragment
+    // version is greater than or equal to 10, we have a footer version of 5.
+    // version is greater than or equal to 7, we have a footer version of 4.
+    // Otherwise, it is version 3.
+    const int frag_version = std::stoi(name.substr(name.find_last_of('_') + 1));
+    if (frag_version >= 10)
+      *version = 5;
+    else if (frag_version >= 7)
+      *version = 4;
+    else
+      *version = 3;
+    return Status::Ok();
+  }
 
-  // The newest version has the 32-byte long UUID at the end
+  // Check if it is in version 1 or 2
+  // Version 2 has the 32-byte long UUID at the end
+  auto t_str = name.substr(name.find_last_of('_') + 1);
   *version = (t_str.size() == 32) ? 2 : 1;
 
   return Status::Ok();
 }
 
+Status get_fragment_version(const std::string& name, uint32_t* version) {
+  uint32_t name_version;
+  RETURN_NOT_OK(get_fragment_name_version(name, &name_version));
+
+  if (name_version <= 2) {
+    *version = UINT32_MAX;
+  } else {  // name version >= 3
+    auto v_str = name.substr(name.find_last_of('_') + 1);
+    std::stringstream ss(v_str);
+    ss >> *version;
+  }
+
+  return Status::Ok();
+}
+
 bool is_int(const std::string& str) {
   // Check if empty
   if (str.empty())
@@ -297,190 +344,6 @@ bool is_uint(const std::string& str) {
   return true;
 }
 
-std::string domain_str(const void* domain, Datatype type) {
-  std::stringstream ss;
-
-  if (domain == nullptr)
-    return "";
-
-  const int* domain_int32;
-  const int64_t* domain_int64;
-  const float* domain_float32;
-  const double* domain_float64;
-  const int8_t* domain_int8;
-  const uint8_t* domain_uint8;
-  const int16_t* domain_int16;
-  const uint16_t* domain_uint16;
-  const uint32_t* domain_uint32;
-  const uint64_t* domain_uint64;
-
-  switch (type) {
-    case Datatype::INT32:
-      domain_int32 = static_cast<const int*>(domain);
-      ss << "[" << domain_int32[0] << "," << domain_int32[1] << "]";
-      return ss.str();
-    case Datatype::INT64:
-      domain_int64 = static_cast<const int64_t*>(domain);
-      ss << "[" << domain_int64[0] << "," << domain_int64[1] << "]";
-      return ss.str();
-    case Datatype::FLOAT32:
-      domain_float32 = static_cast<const float*>(domain);
-      ss << "[" << domain_float32[0] << "," << domain_float32[1] << "]";
-      return ss.str();
-    case Datatype::FLOAT64:
-      domain_float64 = static_cast<const double*>(domain);
-      ss << "[" << domain_float64[0] << "," << domain_float64[1] << "]";
-      return ss.str();
-    case Datatype::INT8:
-      domain_int8 = static_cast<const int8_t*>(domain);
-      ss << "[" << int(domain_int8[0]) << "," << int(domain_int8[1]) << "]";
-      return ss.str();
-    case Datatype::UINT8:
-      domain_uint8 = static_cast<const uint8_t*>(domain);
-      ss << "[" << int(domain_uint8[0]) << "," << int(domain_uint8[1]) << "]";
-      return ss.str();
-    case Datatype::INT16:
-      domain_int16 = static_cast<const int16_t*>(domain);
-      ss << "[" << domain_int16[0] << "," << domain_int16[1] << "]";
-      return ss.str();
-    case Datatype::UINT16:
-      domain_uint16 = static_cast<const uint16_t*>(domain);
-      ss << "[" << domain_uint16[0] << "," << domain_uint16[1] << "]";
-      return ss.str();
-    case Datatype::UINT32:
-      domain_uint32 = static_cast<const uint32_t*>(domain);
-      ss << "[" << domain_uint32[0] << "," << domain_uint32[1] << "]";
-      return ss.str();
-    case Datatype::UINT64:
-      domain_uint64 = static_cast<const uint64_t*>(domain);
-      ss << "[" << domain_uint64[0] << "," << domain_uint64[1] << "]";
-      return ss.str();
-    case Datatype::DATETIME_YEAR:
-    case Datatype::DATETIME_MONTH:
-    case Datatype::DATETIME_WEEK:
-    case Datatype::DATETIME_DAY:
-    case Datatype::DATETIME_HR:
-    case Datatype::DATETIME_MIN:
-    case Datatype::DATETIME_SEC:
-    case Datatype::DATETIME_MS:
-    case Datatype::DATETIME_US:
-    case Datatype::DATETIME_NS:
-    case Datatype::DATETIME_PS:
-    case Datatype::DATETIME_FS:
-    case Datatype::DATETIME_AS:
-      domain_int64 = static_cast<const int64_t*>(domain);
-      ss << "[" << domain_int64[0] << "," << domain_int64[1] << "]";
-      return ss.str();
-
-    case Datatype::CHAR:
-    case Datatype::STRING_ASCII:
-    case Datatype::STRING_UTF8:
-    case Datatype::STRING_UTF16:
-    case Datatype::STRING_UTF32:
-    case Datatype::STRING_UCS2:
-    case Datatype::STRING_UCS4:
-    case Datatype::ANY:
-      // Not supported domain type
-      assert(false);
-      return "";
-  }
-
-  assert(false);
-  return "";
-}
-
-std::string tile_extent_str(const void* tile_extent, Datatype type) {
-  std::stringstream ss;
-
-  if (tile_extent == nullptr)
-    return constants::null_str;
-
-  const int* tile_extent_int32;
-  const int64_t* tile_extent_int64;
-  const float* tile_extent_float32;
-  const double* tile_extent_float64;
-  const int8_t* tile_extent_int8;
-  const uint8_t* tile_extent_uint8;
-  const int16_t* tile_extent_int16;
-  const uint16_t* tile_extent_uint16;
-  const uint32_t* tile_extent_uint32;
-  const uint64_t* tile_extent_uint64;
-
-  switch (type) {
-    case Datatype::INT32:
-      tile_extent_int32 = static_cast<const int*>(tile_extent);
-      ss << *tile_extent_int32;
-      return ss.str();
-    case Datatype::INT64:
-      tile_extent_int64 = static_cast<const int64_t*>(tile_extent);
-      ss << *tile_extent_int64;
-      return ss.str();
-    case Datatype::FLOAT32:
-      tile_extent_float32 = static_cast<const float*>(tile_extent);
-      ss << *tile_extent_float32;
-      return ss.str();
-    case Datatype::FLOAT64:
-      tile_extent_float64 = static_cast<const double*>(tile_extent);
-      ss << *tile_extent_float64;
-      return ss.str();
-    case Datatype::INT8:
-      tile_extent_int8 = static_cast<const int8_t*>(tile_extent);
-      ss << int(*tile_extent_int8);
-      return ss.str();
-    case Datatype::UINT8:
-      tile_extent_uint8 = static_cast<const uint8_t*>(tile_extent);
-      ss << int(*tile_extent_uint8);
-      return ss.str();
-    case Datatype::INT16:
-      tile_extent_int16 = static_cast<const int16_t*>(tile_extent);
-      ss << *tile_extent_int16;
-      return ss.str();
-    case Datatype::UINT16:
-      tile_extent_uint16 = static_cast<const uint16_t*>(tile_extent);
-      ss << *tile_extent_uint16;
-      return ss.str();
-    case Datatype::UINT32:
-      tile_extent_uint32 = static_cast<const uint32_t*>(tile_extent);
-      ss << *tile_extent_uint32;
-      return ss.str();
-    case Datatype::UINT64:
-      tile_extent_uint64 = static_cast<const uint64_t*>(tile_extent);
-      ss << *tile_extent_uint64;
-      return ss.str();
-    case Datatype::DATETIME_YEAR:
-    case Datatype::DATETIME_MONTH:
-    case Datatype::DATETIME_WEEK:
-    case Datatype::DATETIME_DAY:
-    case Datatype::DATETIME_HR:
-    case Datatype::DATETIME_MIN:
-    case Datatype::DATETIME_SEC:
-    case Datatype::DATETIME_MS:
-    case Datatype::DATETIME_US:
-    case Datatype::DATETIME_NS:
-    case Datatype::DATETIME_PS:
-    case Datatype::DATETIME_FS:
-    case Datatype::DATETIME_AS:
-      tile_extent_int64 = static_cast<const int64_t*>(tile_extent);
-      ss << *tile_extent_int64;
-      return ss.str();
-
-    case Datatype::CHAR:
-    case Datatype::STRING_ASCII:
-    case Datatype::STRING_UTF8:
-    case Datatype::STRING_UTF16:
-    case Datatype::STRING_UTF32:
-    case Datatype::STRING_UCS2:
-    case Datatype::STRING_UCS4:
-    case Datatype::ANY:
-      // Not supported domain type
-      assert(false);
-      return "";
-  }
-
-  assert(false);
-  return "";
-}
-
 bool starts_with(const std::string& value, const std::string& prefix) {
   if (prefix.size() > value.size())
     return false;
@@ -501,6 +364,104 @@ std::string to_str(const T& value) {
   return ss.str();
 }
 
+std::string to_str(const void* value, Datatype type) {
+  std::stringstream ss;
+  switch (type) {
+    case Datatype::INT8:
+      ss << *(const int8_t*)value;
+      break;
+    case Datatype::UINT8:
+      ss << *(const uint8_t*)value;
+      break;
+    case Datatype::INT16:
+      ss << *(const int16_t*)value;
+      break;
+    case Datatype::UINT16:
+      ss << *(const uint16_t*)value;
+      break;
+    case Datatype::INT32:
+      ss << *(const int32_t*)value;
+      break;
+    case Datatype::UINT32:
+      ss << *(const uint32_t*)value;
+      break;
+    case Datatype::INT64:
+      ss << *(const int64_t*)value;
+      break;
+    case Datatype::UINT64:
+      ss << *(const uint64_t*)value;
+      break;
+    case Datatype::FLOAT32:
+      ss << *(const float*)value;
+      break;
+    case Datatype::FLOAT64:
+      ss << *(const double*)value;
+      break;
+    case Datatype::CHAR:
+      ss << *(const char*)value;
+      break;
+    case Datatype::ANY:
+      ss << *(const uint8_t*)value;
+      break;
+    case Datatype::STRING_ASCII:
+      ss << *(const uint8_t*)value;
+      break;
+    case Datatype::STRING_UTF8:
+      ss << *(const uint8_t*)value;
+      break;
+    case Datatype::STRING_UTF16:
+      ss << *(const uint16_t*)value;
+      break;
+    case Datatype::STRING_UTF32:
+      ss << *(const uint32_t*)value;
+      break;
+    case Datatype::STRING_UCS2:
+      ss << *(const uint16_t*)value;
+      break;
+    case Datatype::STRING_UCS4:
+      ss << *(const uint32_t*)value;
+      break;
+    case Datatype::DATETIME_YEAR:
+    case Datatype::DATETIME_MONTH:
+    case Datatype::DATETIME_WEEK:
+    case Datatype::DATETIME_DAY:
+    case Datatype::DATETIME_HR:
+    case Datatype::DATETIME_MIN:
+    case Datatype::DATETIME_SEC:
+    case Datatype::DATETIME_MS:
+    case Datatype::DATETIME_US:
+    case Datatype::DATETIME_NS:
+    case Datatype::DATETIME_PS:
+    case Datatype::DATETIME_FS:
+    case Datatype::DATETIME_AS:
+    case Datatype::TIME_HR:
+    case Datatype::TIME_MIN:
+    case Datatype::TIME_SEC:
+    case Datatype::TIME_MS:
+    case Datatype::TIME_US:
+    case Datatype::TIME_NS:
+    case Datatype::TIME_PS:
+    case Datatype::TIME_FS:
+    case Datatype::TIME_AS:
+      ss << *(const int64_t*)value;
+      break;
+    default:
+      assert(false);
+  }
+
+  return ss.str();
+}
+
+uint64_t common_prefix_size(const std::string& a, const std::string& b) {
+  auto size = std::min(a.size(), b.size());
+  for (size_t i = 0; i < size; ++i) {
+    if (a[i] != b[i])
+      return i;
+  }
+
+  return size;
+}
+
 }  // namespace parse
 
 /* ****************************** */
@@ -633,61 +594,6 @@ inline bool coords_in_rect(
   return true;
 }
 
-template <class T>
-inline bool rect_in_rect(
-    const T* rect_a, const T* rect_b, unsigned int dim_num) {
-  for (unsigned int i = 0; i < dim_num; ++i) {
-    if (rect_a[2 * i] < rect_b[2 * i] || rect_a[2 * i] > rect_b[2 * i + 1] ||
-        rect_a[2 * i + 1] < rect_b[2 * i] ||
-        rect_a[2 * i + 1] > rect_b[2 * i + 1])
-      return false;
-  }
-
-  return true;
-}
-
-template <class T>
-void compute_mbr_union(
-    unsigned dim_num, const T* mbrs, uint64_t mbr_num, T* mbr_union) {
-  // Sanity check
-  if (dim_num == 0 || mbr_num == 0)
-    return;
-
-  // Set the first rectangle to the union
-  auto mbr_size = 2 * dim_num * sizeof(T);
-  std::memcpy(mbr_union, mbrs, mbr_size);
-
-  // Expand the union with every other MBR
-  for (uint64_t i = 1; i < mbr_num; ++i)
-    expand_mbr_with_mbr<T>(mbr_union, &mbrs[i * 2 * dim_num], dim_num);
-}
-
-template <class T>
-void expand_mbr(T* mbr, const T* coords, unsigned int dim_num) {
-  for (unsigned int i = 0; i < dim_num; ++i) {
-    // Update lower bound on dimension i
-    if (mbr[2 * i] > coords[i])
-      mbr[2 * i] = coords[i];
-
-    // Update upper bound on dimension i
-    if (mbr[2 * i + 1] < coords[i])
-      mbr[2 * i + 1] = coords[i];
-  }
-}
-
-template <class T>
-void expand_mbr_with_mbr(T* mbr_a, const T* mbr_b, unsigned int dim_num) {
-  for (unsigned int i = 0; i < dim_num; ++i) {
-    // Update lower bound on dimension i
-    if (mbr_a[2 * i] > mbr_b[2 * i])
-      mbr_a[2 * i] = mbr_b[2 * i];
-
-    // Update upper bound on dimension i
-    if (mbr_a[2 * i + 1] < mbr_b[2 * i + 1])
-      mbr_a[2 * i + 1] = mbr_b[2 * i + 1];
-  }
-}
-
 template <class T>
 bool overlap(const T* a, const T* b, unsigned dim_num) {
   for (unsigned i = 0; i < dim_num; ++i) {
@@ -698,31 +604,13 @@ bool overlap(const T* a, const T* b, unsigned dim_num) {
   return true;
 }
 
-template <class T>
-bool overlap(const T* a, const T* b, unsigned dim_num, bool* a_contains_b) {
-  for (unsigned i = 0; i < dim_num; ++i) {
-    if (a[2 * i] > b[2 * i + 1] || a[2 * i + 1] < b[2 * i])
-      return false;
-  }
-
-  *a_contains_b = true;
-  for (unsigned i = 0; i < dim_num; ++i) {
-    if (a[2 * i] > b[2 * i] || a[2 * i + 1] < b[2 * i + 1]) {
-      *a_contains_b = false;
-      break;
-    }
-  }
-
-  return true;
-}
-
 template <class T>
 void overlap(const T* a, const T* b, unsigned dim_num, T* o, bool* overlap) {
   // Get overlap range
   *overlap = true;
   for (unsigned int i = 0; i < dim_num; ++i) {
-    o[2 * i] = MAX(a[2 * i], b[2 * i]);
-    o[2 * i + 1] = MIN(a[2 * i + 1], b[2 * i + 1]);
+    o[2 * i] = std::max(a[2 * i], b[2 * i]);
+    o[2 * i + 1] = std::min(a[2 * i + 1], b[2 * i + 1]);
     if (o[2 * i] > b[2 * i + 1] || o[2 * i + 1] < b[2 * i]) {
       *overlap = false;
       break;
@@ -730,17 +618,6 @@ void overlap(const T* a, const T* b, unsigned dim_num, T* o, bool* overlap) {
   }
 }
 
-template <class T>
-bool overlap(const std::vector<const T*>& a, const T* b) {
-  auto dim_num = (unsigned)a.size();
-  for (unsigned i = 0; i < dim_num; ++i) {
-    if (a[i][0] > b[2 * i + 1] || a[i][1] < b[2 * i])
-      return false;
-  }
-
-  return true;
-}
-
 template <class T>
 double coverage(const T* a, const T* b, unsigned dim_num) {
   double c = 1.0;
@@ -752,7 +629,7 @@ double coverage(const T* a, const T* b, unsigned dim_num) {
     } else {
       auto a_range = double(a[2 * i + 1]) - a[2 * i] + add;
       auto b_range = double(b[2 * i + 1]) - b[2 * i] + add;
-      if (std::numeric_limits<T>::is_integer) {
+      if (std::is_integral<T>::value) {
         auto max = std::numeric_limits<T>::max();
         if (a_range == 0)
           a_range = std::nextafter(a_range, max);
@@ -765,6 +642,20 @@ double coverage(const T* a, const T* b, unsigned dim_num) {
   return c;
 }
 
+template <class T>
+std::vector<std::array<T, 2>> intersection(
+    const std::vector<std::array<T, 2>>& r1,
+    const std::vector<std::array<T, 2>>& r2) {
+  auto dim_num = r1.size();
+  assert(r2.size() == dim_num);
+
+  std::vector<std::array<T, 2>> ret(dim_num);
+  for (size_t d = 0; d < dim_num; ++d)
+    ret[d] = {std::max(r1[d][0], r2[d][0]), std::min(r1[d][1], r2[d][1])};
+
+  return ret;
+}
+
 }  // namespace geometry
 
 /* ********************************* */
@@ -819,6 +710,32 @@ T safe_mul(T a, T b) {
   return prod;
 }
 
+uint64_t left_p2_m1(uint64_t value) {
+  // Edge case
+  if (value == UINT64_MAX)
+    return value;
+
+  uint64_t ret = 0;  // Min power of 2 minus 1
+  while (ret <= value) {
+    ret = (ret << 1) | 1;  // Next larger power of 2 minus 1
+  }
+
+  return ret >> 1;
+}
+
+uint64_t right_p2_m1(uint64_t value) {
+  // Edge case
+  if (value == 0)
+    return value;
+
+  uint64_t ret = UINT64_MAX;  // Max power of 2 minus 1
+  while (ret >= value) {
+    ret >>= 1;  // Next smaller power of 2 minus 1
+  }
+
+  return (ret << 1) | 1;
+}
+
 }  // namespace math
 
 // Explicit template instantiations
@@ -887,69 +804,6 @@ template bool coords_in_rect<uint64_t>(
     const std::vector<const uint64_t*>& subarray,
     unsigned int dim_num);
 
-template bool rect_in_rect<int>(
-    const int* rect_a, const int* rect_b, unsigned int dim_num);
-template bool rect_in_rect<int64_t>(
-    const int64_t* rect_a, const int64_t* rect_b, unsigned int dim_num);
-template bool rect_in_rect<float>(
-    const float* react_a, const float* rect_b, unsigned int dim_num);
-template bool rect_in_rect<double>(
-    const double* rect_a, const double* rect_b, unsigned int dim_num);
-template bool rect_in_rect<int8_t>(
-    const int8_t* rect_a, const int8_t* rect_b, unsigned int dim_num);
-template bool rect_in_rect<uint8_t>(
-    const uint8_t* rect_a, const uint8_t* rect_b, unsigned int dim_num);
-template bool rect_in_rect<int16_t>(
-    const int16_t* rect_a, const int16_t* rect_b, unsigned int dim_num);
-template bool rect_in_rect<uint16_t>(
-    const uint16_t* rect_a, const uint16_t* rect_b, unsigned int dim_num);
-template bool rect_in_rect<uint32_t>(
-    const uint32_t* rect_a, const uint32_t* rect_b, unsigned int dim_num);
-template bool rect_in_rect<uint64_t>(
-    const uint64_t* rect_a, const uint64_t* rect_b, unsigned int dim_num);
-
-template void expand_mbr<int>(
-    int* mbr, const int* coords, unsigned int dim_num);
-template void expand_mbr<int64_t>(
-    int64_t* mbr, const int64_t* coords, unsigned int dim_num);
-template void expand_mbr<float>(
-    float* mbr, const float* coords, unsigned int dim_num);
-template void expand_mbr<double>(
-    double* mbr, const double* coords, unsigned int dim_num);
-template void expand_mbr<int8_t>(
-    int8_t* mbr, const int8_t* coords, unsigned int dim_num);
-template void expand_mbr<uint8_t>(
-    uint8_t* mbr, const uint8_t* coords, unsigned int dim_num);
-template void expand_mbr<int16_t>(
-    int16_t* mbr, const int16_t* coords, unsigned int dim_num);
-template void expand_mbr<uint16_t>(
-    uint16_t* mbr, const uint16_t* coords, unsigned int dim_num);
-template void expand_mbr<uint32_t>(
-    uint32_t* mbr, const uint32_t* coords, unsigned int dim_num);
-template void expand_mbr<uint64_t>(
-    uint64_t* mbr, const uint64_t* coords, unsigned int dim_num);
-
-template void expand_mbr_with_mbr<int>(
-    int* mbr_a, const int* mbr_b, unsigned int dim_num);
-template void expand_mbr_with_mbr<int64_t>(
-    int64_t* mbr_a, const int64_t* mbr_b, unsigned int dim_num);
-template void expand_mbr_with_mbr<float>(
-    float* mbr_a, const float* mbr_b, unsigned int dim_num);
-template void expand_mbr_with_mbr<double>(
-    double* mbr_a, const double* mbr_b, unsigned int dim_num);
-template void expand_mbr_with_mbr<int8_t>(
-    int8_t* mbr_a, const int8_t* mbr_b, unsigned int dim_num);
-template void expand_mbr_with_mbr<uint8_t>(
-    uint8_t* mbr_a, const uint8_t* mbr_b, unsigned int dim_num);
-template void expand_mbr_with_mbr<int16_t>(
-    int16_t* mbr_a, const int16_t* mbr_b, unsigned int dim_num);
-template void expand_mbr_with_mbr<uint16_t>(
-    uint16_t* mbr_a, const uint16_t* mbr_b, unsigned int dim_num);
-template void expand_mbr_with_mbr<uint32_t>(
-    uint32_t* mbr_a, const uint32_t* mbr_b, unsigned int dim_num);
-template void expand_mbr_with_mbr<uint64_t>(
-    uint64_t* mbr_a, const uint64_t* mbr_b, unsigned int dim_num);
-
 template bool overlap<int8_t>(
     const int8_t* a, const int8_t* b, unsigned dim_num);
 template bool overlap<uint8_t>(
@@ -969,27 +823,6 @@ template bool overlap<float>(const float* a, const float* b, unsigned dim_num);
 template bool overlap<double>(
     const double* a, const double* b, unsigned dim_num);
 
-template bool overlap<int8_t>(
-    const int8_t* a, const int8_t* b, unsigned dim_num, bool* a_contains_b);
-template bool overlap<uint8_t>(
-    const uint8_t* a, const uint8_t* b, unsigned dim_num, bool* a_contains_b);
-template bool overlap<int16_t>(
-    const int16_t* a, const int16_t* b, unsigned dim_num, bool* a_contains_b);
-template bool overlap<uint16_t>(
-    const uint16_t* a, const uint16_t* b, unsigned dim_num, bool* a_contains_b);
-template bool overlap<int>(
-    const int* a, const int* b, unsigned dim_num, bool* a_contains_b);
-template bool overlap<unsigned>(
-    const unsigned* a, const unsigned* b, unsigned dim_num, bool* a_contains_b);
-template bool overlap<int64_t>(
-    const int64_t* a, const int64_t* b, unsigned dim_num, bool* a_contains_b);
-template bool overlap<uint64_t>(
-    const uint64_t* a, const uint64_t* b, unsigned dim_num, bool* a_contains_b);
-template bool overlap<float>(
-    const float* a, const float* b, unsigned dim_num, bool* a_contains_b);
-template bool overlap<double>(
-    const double* a, const double* b, unsigned dim_num, bool* a_contains_b);
-
 template void overlap<int>(
     const int* a, const int* b, unsigned dim_num, int* o, bool* overlap);
 template void overlap<int64_t>(
@@ -1043,27 +876,6 @@ template void overlap<uint64_t>(
     uint64_t* o,
     bool* overlap);
 
-template bool overlap<int8_t>(
-    const std::vector<const int8_t*>& a, const int8_t* b);
-template bool overlap<uint8_t>(
-    const std::vector<const uint8_t*>& a, const uint8_t* b);
-template bool overlap<int16_t>(
-    const std::vector<const int16_t*>& a, const int16_t* b);
-template bool overlap<uint16_t>(
-    const std::vector<const uint16_t*>& a, const uint16_t* b);
-template bool overlap<int32_t>(
-    const std::vector<const int32_t*>& a, const int32_t* b);
-template bool overlap<uint32_t>(
-    const std::vector<const uint32_t*>& a, const uint32_t* b);
-template bool overlap<int64_t>(
-    const std::vector<const int64_t*>& a, const int64_t* b);
-template bool overlap<uint64_t>(
-    const std::vector<const uint64_t*>& a, const uint64_t* b);
-template bool overlap<float>(
-    const std::vector<const float*>& a, const float* b);
-template bool overlap<double>(
-    const std::vector<const double*>& a, const double* b);
-
 template double coverage<int8_t>(
     const int8_t* a, const int8_t* b, unsigned dim_num);
 template double coverage<uint8_t>(
@@ -1084,47 +896,30 @@ template double coverage<float>(
 template double coverage<double>(
     const double* a, const double* b, unsigned dim_num);
 
-template void compute_mbr_union<int8_t>(
-    unsigned dim_num, const int8_t* mbrs, uint64_t mbr_num, int8_t* mbr_union);
-template void compute_mbr_union<uint8_t>(
-    unsigned dim_num,
-    const uint8_t* mbrs,
-    uint64_t mbr_num,
-    uint8_t* mbr_union);
-template void compute_mbr_union<int16_t>(
-    unsigned dim_num,
-    const int16_t* mbrs,
-    uint64_t mbr_num,
-    int16_t* mbr_union);
-template void compute_mbr_union<uint16_t>(
-    unsigned dim_num,
-    const uint16_t* mbrs,
-    uint64_t mbr_num,
-    uint16_t* mbr_union);
-template void compute_mbr_union<int32_t>(
-    unsigned dim_num,
-    const int32_t* mbrs,
-    uint64_t mbr_num,
-    int32_t* mbr_union);
-template void compute_mbr_union<uint32_t>(
-    unsigned dim_num,
-    const uint32_t* mbrs,
-    uint64_t mbr_num,
-    uint32_t* mbr_union);
-template void compute_mbr_union<int64_t>(
-    unsigned dim_num,
-    const int64_t* mbrs,
-    uint64_t mbr_num,
-    int64_t* mbr_union);
-template void compute_mbr_union<uint64_t>(
-    unsigned dim_num,
-    const uint64_t* mbrs,
-    uint64_t mbr_num,
-    uint64_t* mbr_union);
-template void compute_mbr_union<float>(
-    unsigned dim_num, const float* mbrs, uint64_t mbr_num, float* mbr_union);
-template void compute_mbr_union<double>(
-    unsigned dim_num, const double* mbrs, uint64_t mbr_num, double* mbr_union);
+template std::vector<std::array<int8_t, 2>> intersection<int8_t>(
+    const std::vector<std::array<int8_t, 2>>& r1,
+    const std::vector<std::array<int8_t, 2>>& r2);
+template std::vector<std::array<uint8_t, 2>> intersection<uint8_t>(
+    const std::vector<std::array<uint8_t, 2>>& r1,
+    const std::vector<std::array<uint8_t, 2>>& r2);
+template std::vector<std::array<int16_t, 2>> intersection<int16_t>(
+    const std::vector<std::array<int16_t, 2>>& r1,
+    const std::vector<std::array<int16_t, 2>>& r2);
+template std::vector<std::array<uint16_t, 2>> intersection<uint16_t>(
+    const std::vector<std::array<uint16_t, 2>>& r1,
+    const std::vector<std::array<uint16_t, 2>>& r2);
+template std::vector<std::array<int32_t, 2>> intersection<int32_t>(
+    const std::vector<std::array<int32_t, 2>>& r1,
+    const std::vector<std::array<int32_t, 2>>& r2);
+template std::vector<std::array<uint32_t, 2>> intersection<uint32_t>(
+    const std::vector<std::array<uint32_t, 2>>& r1,
+    const std::vector<std::array<uint32_t, 2>>& r2);
+template std::vector<std::array<int64_t, 2>> intersection<int64_t>(
+    const std::vector<std::array<int64_t, 2>>& r1,
+    const std::vector<std::array<int64_t, 2>>& r2);
+template std::vector<std::array<uint64_t, 2>> intersection<uint64_t>(
+    const std::vector<std::array<uint64_t, 2>>& r1,
+    const std::vector<std::array<uint64_t, 2>>& r2);
 
 }  // namespace geometry
 
