@@ -5,7 +5,7 @@
  *
  * The MIT License
  *
- * @copyright Copyright (c) 2017-2019 TileDB, Inc.
+ * @copyright Copyright (c) 2017-2021 TileDB, Inc.
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
  * of this software and associated documentation files (the "Software"), to deal
@@ -33,26 +33,30 @@
 #ifndef TILEDB_WRITER_H
 #define TILEDB_WRITER_H
 
-#include "tiledb/sm/array_schema/array_schema.h"
-#include "tiledb/sm/filter/filter_pipeline.h"
-#include "tiledb/sm/fragment/fragment_metadata.h"
+#include <atomic>
+
+#include "tiledb/common/status.h"
 #include "tiledb/sm/fragment/written_fragment_info.h"
-#include "tiledb/sm/misc/status.h"
-#include "tiledb/sm/query/dense_cell_range_iter.h"
-#include "tiledb/sm/query/types.h"
+#include "tiledb/sm/misc/types.h"
+#include "tiledb/sm/query/dense_tiler.h"
+#include "tiledb/sm/query/iquery_strategy.h"
+#include "tiledb/sm/query/query.h"
+#include "tiledb/sm/query/query_buffer.h"
+#include "tiledb/sm/query/strategy_base.h"
+#include "tiledb/sm/stats/stats.h"
 #include "tiledb/sm/tile/tile.h"
 
-#include <memory>
-#include <set>
+using namespace tiledb::common;
 
 namespace tiledb {
 namespace sm {
 
 class Array;
+class FragmentMetadata;
 class StorageManager;
 
 /** Processes write queries. */
-class Writer {
+class Writer : public StrategyBase, public IQueryStrategy {
  public:
   /* ********************************* */
   /*          TYPE DEFINITIONS         */
@@ -64,148 +68,63 @@ class Writer {
    */
   struct GlobalWriteState {
     /**
-     * Stores the last tile of each attribute for each write operation.
-     * For fixed-sized attributes, the second tile is ignored. For
-     * var-sized attributes, the first tile is the offsets tile, whereas
-     * the second tile is the values tile.
+     * Stores the last tile of each attribute/dimension for each write
+     * operation. For fixed-sized attributes/dimensions, the second tile is
+     * ignored. For var-sized attributes/dimensions, the first tile is the
+     * offsets tile, whereas the second tile is the values tile. In both cases,
+     * the third tile stores a validity tile for nullable attributes.
      */
-    std::unordered_map<std::string, std::pair<Tile, Tile>> last_tiles_;
+    std::unordered_map<std::string, std::tuple<Tile, Tile, Tile>> last_tiles_;
 
     /**
-     * Stores the number of cells written for each attribute across the
-     * write operations.
+     * Stores the number of cells written for each attribute/dimension across
+     * the write operations.
      */
     std::unordered_map<std::string, uint64_t> cells_written_;
 
-    /** The fragment metadata. */
-    std::shared_ptr<FragmentMetadata> frag_meta_;
+    /** The fragment metadata that the writer will focus on. */
+    tdb_shared_ptr<FragmentMetadata> frag_meta_;
   };
 
-  /** Cell range to be written. */
-  struct WriteCellRange {
-    /** The position in the tile where the range will be copied. */
-    uint64_t pos_;
-    /** The starting cell in the user buffers. */
-    uint64_t start_;
-    /** The ending cell in the user buffers. */
-    uint64_t end_;
-
-    /** Constructor. */
-    WriteCellRange(uint64_t pos, uint64_t start, uint64_t end)
-        : pos_(pos)
-        , start_(start)
-        , end_(end) {
-    }
-  };
-
-  /** A vector of write cell ranges. */
-  typedef std::vector<WriteCellRange> WriteCellRangeVec;
-
   /* ********************************* */
   /*     CONSTRUCTORS & DESTRUCTORS    */
   /* ********************************* */
 
   /** Constructor. */
-  Writer();
+  Writer(
+      stats::Stats* stats,
+      tdb_shared_ptr<Logger> logger,
+      StorageManager* storage_manager,
+      Array* array,
+      Config& config,
+      std::unordered_map<std::string, QueryBuffer>& buffers,
+      Subarray& subarray,
+      Layout layout,
+      std::vector<WrittenFragmentInfo>& written_fragment_info,
+      bool disable_check_global_order,
+      Query::CoordsInfo& coords_info_,
+      URI fragment_uri = URI(""));
 
   /** Destructor. */
   ~Writer();
 
+  DISABLE_COPY_AND_COPY_ASSIGN(Writer);
+  DISABLE_MOVE_AND_MOVE_ASSIGN(Writer);
+
   /* ********************************* */
   /*                 API               */
   /* ********************************* */
 
-  /**
-   * Adds a range to the (read/write) query on the input dimension,
-   * in the form of (start, end, stride).
-   * The range components must be of the same type as the domain type of the
-   * underlying array.
-   */
-  Status add_range(
-      unsigned dim_idx, const void* start, const void* end, const void* stride);
+  /** Returns the names of the buffers set by the user for the write query. */
+  std::vector<std::string> buffer_names() const;
 
-  /** Retrieves the number of ranges of the subarray for the given dimension. */
-  Status get_range_num(unsigned dim_idx, uint64_t* range_num) const;
-
-  /**
-   * Retrieves a range from a dimension in the form (start, end, stride).
-   *
-   * @param dim_idx The dimension to retrieve the range from.
-   * @param range_idx The id of the range to retrieve.
-   * @param start The range start to retrieve.
-   * @param end The range end to retrieve.
-   * @param stride The range stride to retrieve.
-   * @return Status
-   */
-  Status get_range(
-      unsigned dim_idx,
-      uint64_t range_idx,
-      const void** start,
-      const void** end,
-      const void** stride) const;
-
-  /**
-   * Gets the estimated result size (in bytes) for the input fixed-sized
-   * attribute.
-   */
-  Status get_est_result_size(const char* attr_name, uint64_t* size);
-
-  /**
-   * Gets the estimated result size (in bytes) for the input var-sized
-   * attribute.
-   */
-  Status get_est_result_size(
-      const char* attr_name, uint64_t* size_off, uint64_t* size_val);
-
-  /** Returns the array schema. */
-  const ArraySchema* array_schema() const;
-
-  /**
-   * Return list of attribtues for query
-   * @return vector of attributes for query
-   */
-  std::vector<std::string> attributes() const;
-
-  /**
-   * Fetch AttributeBuffer for attribute
-   * @param attribute to fetch
-   * @return AttributeBuffer for attribute
-   */
-  AttributeBuffer buffer(const std::string& attribute) const;
-
-  /** Finalizes the reader. */
+  /** Finalizes the writer. */
   Status finalize();
 
-  /**
-   * Retrieves the buffer of a fixed-sized attribute.
-   *
-   * @param attribute The buffer attribute.
-   * @param buffer The buffer to be retrieved.
-   * @param buffer_size A pointer to the buffer size to be retrieved.
-   * @return Status
-   */
-  Status get_buffer(
-      const std::string& attribute,
-      void** buffer,
-      uint64_t** buffer_size) const;
-
-  /**
-   * Retrieves the offsets and values buffers of a var-sized attribute.
-   *
-   * @param attribute The buffer attribute.
-   * @param buffer_off The offsets buffer to be retrieved.
-   * @param buffer_off_size A pointer to the offsets buffer size to be
-   * retrieved.
-   * @param buffer_val The values buffer to be retrieved.
-   * @param buffer_val_size A pointer to the values buffer size to be retrieved.
-   * @return Status
-   */
-  Status get_buffer(
-      const std::string& attribute,
-      uint64_t** buffer_off,
-      uint64_t** buffer_off_size,
-      void** buffer_val,
-      uint64_t** buffer_val_size) const;
+  /** Writer is never in an imcomplete state. */
+  bool incomplete() const {
+    return false;
+  }
 
   /** Returns current setting of check_coord_dups_ */
   bool get_check_coord_dups() const;
@@ -219,48 +138,8 @@ class Writer {
   /** Initializes the writer. */
   Status init();
 
-  /** Returns the cell layout. */
-  Layout layout() const;
-
-  /** Sets the array. */
-  void set_array(const Array* array);
-
-  /*
-   * Sets the array schema. If the array is a kv store, then this
-   * function also sets global order as the default layout.
-   */
-  void set_array_schema(const ArraySchema* array_schema);
-
-  /**
-   * Sets the buffer for a fixed-sized attribute.
-   *
-   * @param attribute The attribute to set the buffer for.
-   * @param buffer The buffer that has the input data to be written.
-   * @param buffer_size The size of `buffer` in bytes.
-   * @return Status
-   */
-  Status set_buffer(
-      const std::string& attribute, void* buffer, uint64_t* buffer_size);
-
-  /**
-   * Sets the buffer for a var-sized attribute.
-   *
-   * @param attribute The attribute to set the buffer for.
-   * @param buffer_off The buffer that has the input data to be written,
-   *     This buffer holds the starting offsets of each cell value in
-   *     `buffer_val`.
-   * @param buffer_off_size The size of `buffer_off` in bytes.
-   * @param buffer_val The buffer that has the input data to be written.
-   *     This buffer holds the actual var-sized cell values.
-   * @param buffer_val_size The size of `buffer_val` in bytes.
-   * @return Status
-   */
-  Status set_buffer(
-      const std::string& attribute,
-      uint64_t* buffer_off,
-      uint64_t* buffer_off_size,
-      void* buffer_val,
-      uint64_t* buffer_val_size);
+  /** Initialize the memory budget variables. */
+  Status initialize_memory_budget();
 
   /** Sets current setting of check_coord_dups_ */
   void set_check_coord_dups(bool b);
@@ -271,64 +150,29 @@ class Writer {
   /** Sets current setting of dedup_coords_ */
   void set_dedup_coords(bool b);
 
-  /** Sets the fragment URI. Applicable only to write queries. */
-  void set_fragment_uri(const URI& fragment_uri);
-
-  /**
-   * Sets the cell layout of the query. The function will return an error
-   * if the queried array is a key-value store (because it has its default
-   * layout for both reads and writes.
-   */
-  Status set_layout(Layout layout);
-
-  /** Sets the storage manager. */
-  void set_storage_manager(StorageManager* storage_manager);
-
-  /**
-   * Sets the query subarray. If it is null, then the subarray will be set to
-   * the entire domain.
-   *
-   * @param subarray The subarray to be set.
-   * @return Status
-   */
-  Status set_subarray(const void* subarray);
-
-  /**
-   * Sets the query subarray.
-   *
-   * @param subarray The subarray to be set.
-   * @return Status
-   */
-  Status set_subarray(const Subarray& subarray);
-
-  /*
-   * Return the subarray
-   * @return subarray
-   */
-  void* subarray() const;
-
   /** Performs a write query using its set members. */
-  Status write();
-
-  /** Returns the written fragment info. */
-  const std::vector<WrittenFragmentInfo>& written_fragment_info() const;
+  Status dowork();
 
  private:
   /* ********************************* */
   /*         PRIVATE ATTRIBUTES        */
   /* ********************************* */
 
-  /** The array. */
-  const Array* array_;
+  /**
+   * The sizes of the coordinate buffers in a map (dimension -> size).
+   * Needed separate storage since QueryBuffer stores a pointer to the buffer
+   * sizes.
+   */
+  std::unordered_map<std::string, uint64_t> coord_buffer_sizes_;
 
-  /** The array schema. */
-  const ArraySchema* array_schema_;
+  /**
+   * If `true`, it will not check if the written coordinates are
+   * in the global order. This supercedes the config.
+   */
+  bool disable_check_global_order_;
 
-  /** The names of the attributes involved in the query. */
-  std::vector<std::string> attributes_;
-
-  /** Maps attribute names to their buffers. */
-  std::unordered_map<std::string, AttributeBuffer> attr_buffers_;
+  /** Keeps track of the coords data. */
+  Query::CoordsInfo& coords_info_;
 
   /**
    * Meaningful only when `dedup_coords_` is `false`.
@@ -363,36 +207,26 @@ class Writer {
   URI fragment_uri_;
 
   /** The state associated with global writes. */
-  std::unique_ptr<GlobalWriteState> global_write_state_;
+  tdb_unique_ptr<GlobalWriteState> global_write_state_;
 
   /** True if the writer has been initialized. */
   bool initialized_;
 
-  /**
-   * The layout of the cells in the result of the subarray. Note
-   * that this may not be the same as what the user set to the
-   * query, as the Writer may calibrate it to boost performance.
-   */
-  Layout layout_;
-
-  /** The storage manager. */
-  StorageManager* storage_manager_;
-
-  /** The subarray the query is constrained on. */
-  void* subarray_;
-
   /** Stores information about the written fragments. */
-  std::vector<WrittenFragmentInfo> written_fragment_info_;
+  std::vector<WrittenFragmentInfo>& written_fragment_info_;
+
+  /** Allocated buffers that neeed to be cleaned upon destruction. */
+  std::vector<void*> to_clean_;
+
+  /** UID of the logger instance */
+  inline static std::atomic<uint64_t> logger_id_ = 0;
 
   /* ********************************* */
   /*           PRIVATE METHODS         */
   /* ********************************* */
 
   /** Adss a fragment to `written_fragment_info_`. */
-  void add_written_fragment_info(const URI& uri);
-
-  /** Checks if attributes has been appropriately set for the query. */
-  Status check_attributes();
+  Status add_written_fragment_info(const URI& uri);
 
   /** Correctness checks for buffer sizes. */
   Status check_buffer_sizes() const;
@@ -414,16 +248,6 @@ class Writer {
    */
   Status check_coord_oob() const;
 
-  /**
-   * Throws an error if there are coordinates falling out-of-bounds, i.e.,
-   * outside the array domain.
-   *
-   * @tparam T The coordinates type.
-   * @return Status
-   */
-  template <class T>
-  Status check_coord_oob() const;
-
   /**
    * Throws an error if there are coordinate duplicates. This function
    * assumes that the coordinates are written in the global layout,
@@ -443,23 +267,30 @@ class Writer {
 
   /**
    * Throws an error if there are coordinates that do not obey the
-   * global order.
+   * global order. Applicable only to Hilbert order.
    *
-   * @tparam T The domain type.
    * @return Status
    */
-  template <class T>
-  Status check_global_order() const;
+  Status check_global_order_hilbert() const;
 
   /** Correctness checks for `subarray_`. */
   Status check_subarray() const;
 
-  /** Correctness checks for `subarray_`. */
-  template <class T>
-  Status check_subarray() const;
+  /**
+   * Check the validity of the provided buffer offsets for a variable attribute.
+   *
+   * @return Status
+   */
+  Status check_var_attr_offsets() const;
+
+  /**
+   * Cleans up the coordinate buffers. Applicable only if the coordinate
+   * buffers were allocated by TileDB (not the user)
+   */
+  void clear_coord_buffers();
 
   /** Closes all attribute files, flushing their state to storage. */
-  Status close_files(FragmentMetadata* meta) const;
+  Status close_files(tdb_shared_ptr<FragmentMetadata> meta) const;
 
   /**
    * Computes the positions of the coordinate duplicates (if any). Note
@@ -497,28 +328,14 @@ class Writer {
   /**
    * Computes the coordinates metadata (e.g., MBRs).
    *
-   * @tparam T The domain type.
-   * @param tiles The tiles to calculate the coords metadata from.
+   * @param tiles The tiles to calculate the coords metadata from. It is
+   *     a vector of vectors, one vector of tiles per dimension.
    * @param meta The fragment metadata that will store the coords metadata.
    * @return Status
    */
-  template <class T>
   Status compute_coords_metadata(
-      const std::vector<Tile>& tiles, FragmentMetadata* meta) const;
-
-  /**
-   * Computes the cell ranges to be written, derived from a
-   * dense cell range iterator for a specific tile.
-   *
-   * @tparam T The domain type.
-   * @param iter The dense cell range iterator for one
-   *     tile overlapping with the write subarray.
-   * @param write_cell_ranges The write cell ranges to be created.
-   * @return Status
-   */
-  template <class T>
-  Status compute_write_cell_ranges(
-      DenseCellRangeIter<T>* iters, WriteCellRangeVec* write_cell_ranges) const;
+      const std::unordered_map<std::string, std::vector<Tile>>& tiles,
+      tdb_shared_ptr<FragmentMetadata> meta) const;
 
   /**
    * Creates a new fragment.
@@ -528,136 +345,143 @@ class Writer {
    * @return Status
    */
   Status create_fragment(
-      bool dense, std::shared_ptr<FragmentMetadata>* frag_meta) const;
+      bool dense, tdb_shared_ptr<FragmentMetadata>& frag_meta) const;
+
+  /**
+   * Runs the input coordinate and attribute tiles through their
+   * filter pipelines. The tile buffers are modified to contain the output
+   * of the pipeline.
+   */
+  Status filter_tiles(
+      std::unordered_map<std::string, std::vector<Tile>>* tiles);
+
+  /**
+   * Applicable only to global writes. Filters the last attribute and
+   * coordinate tiles.
+   */
+  Status filter_last_tiles(
+      std::unordered_map<std::string, std::vector<Tile>>* tiles);
 
   /**
    * Runs the input tiles for the input attribute through the filter pipeline.
    * The tile buffers are modified to contain the output of the pipeline.
    *
-   * @param attribute The attribute the tiles belong to.
+   * @param name The attribute/dimension the tiles belong to.
    * @param tile The tiles to be filtered.
    * @return Status
    */
-  Status filter_tiles(
-      const std::string& attribute, std::vector<Tile>* tiles) const;
+  Status filter_tiles(const std::string& name, std::vector<Tile>* tiles);
 
   /**
-   * Runs the input tile for the input attribute through the filter pipeline.
-   * The tile buffer is modified to contain the output of the pipeline.
+   * Runs the input tile for the input attribute/dimension through the filter
+   * pipeline. The tile buffer is modified to contain the output of the
+   * pipeline.
    *
-   * @param attribute The attribute the tile belong to.
+   * @param name The attribute/dimension the tile belong to.
    * @param tile The tile to be filtered.
    * @param offsets True if the tile to be filtered contains offsets for a
-   *    var-sized attribute.
+   *    var-sized attribute/dimension.
+   * @param offsets True if the tile to be filtered contains validity values.
    * @return Status
    */
   Status filter_tile(
-      const std::string& attribute, Tile* tile, bool offsets) const;
+      const std::string& name, Tile* tile, bool offsets, bool nullable);
 
   /** Finalizes the global write state. */
   Status finalize_global_write_state();
 
-  /**
-   * Finalizes the global write state.
-   *
-   * @tparam T The domain type.
-   * @return Status
-   */
-  template <class T>
-  Status finalize_global_write_state();
-
   /**
    * Writes in the global layout. Applicable to both dense and sparse
    * arrays.
    */
   Status global_write();
 
-  /**
-   * Writes in the global layout. Applicable to both dense and sparse
-   * arrays.
-   *
-   * @tparam T The domain type.
-   */
-  template <class T>
-  Status global_write();
-
   /**
    * Applicable only to global writes. Writes the last tiles for each
    * attribute remaining in the state, and records the metadata for
-   * the coordinates attribute (if present).
+   * the coordinates (if present).
    *
-   * @tparam T The domain type.
    * @return Status
    */
-  template <class T>
   Status global_write_handle_last_tile();
 
-  /** Returns `true` if the coordinates are included in the attributes. */
-  bool has_coords() const;
-
   /** Initializes the global write state. */
   Status init_global_write_state();
 
   /**
    * Initializes a fixed-sized tile.
    *
-   * @param attribute The attribute the tile belongs to.
+   * @param name The attribute/dimension the tile belongs to.
    * @param tile The tile to be initialized.
    * @return Status
    */
-  Status init_tile(const std::string& attribute, Tile* tile) const;
+  Status init_tile(const std::string& name, Tile* tile) const;
 
   /**
    * Initializes a var-sized tile.
    *
-   * @param attribute The attribute the tile belongs to.
+   * @param name The attribute/dimension the tile belongs to.
    * @param tile The offsets tile to be initialized.
    * @param tile_var The var-sized data tile to be initialized.
    * @return Status
    */
-  Status init_tile(
-      const std::string& attribute, Tile* tile, Tile* tile_var) const;
+  Status init_tile(const std::string& name, Tile* tile, Tile* tile_var) const;
 
   /**
-   * Initializes dense cell range iterators for the subarray to be writte,
-   * one per overlapping tile.
+   * Initializes a fixed-sized, nullable tile.
    *
-   * @tparam T The domain type.
-   * @param iters The dense cell range iterators to be created.
+   * @param name The attribute the tile belongs to.
+   * @param tile The tile to be initialized.
+   * @param tile_validity The validity tile to be initialized.
    * @return Status
    */
-  template <class T>
-  Status init_tile_dense_cell_range_iters(
-      std::vector<DenseCellRangeIter<T>>* iters) const;
+  Status init_tile_nullable(
+      const std::string& name, Tile* tile, Tile* tile_validity) const;
 
   /**
-   * Initializes the tiles for writing for the input attribute.
+   * Initializes a var-sized, nullable tile.
    *
-   * @param attribute The attribute the tiles belong to.
+   * @param name The attribute the tile belongs to.
+   * @param tile The offsets tile to be initialized.
+   * @param tile_var The var-sized data tile to be initialized.
+   * @param tile_validity The validity tile to be initialized.
+   * @return Status
+   */
+  Status init_tile_nullable(
+      const std::string& name,
+      Tile* tile,
+      Tile* tile_var,
+      Tile* tile_validity) const;
+
+  /**
+   * Initializes the tiles for writing for the input attribute/dimension.
+   *
+   * @param name The attribute/dimension the tiles belong to.
    * @param tile_num The number of tiles.
    * @param tiles The tiles to be initialized. Note that the vector
    *     has been already preallocated.
    * @return Status
    */
   Status init_tiles(
-      const std::string& attribute,
+      const std::string& name,
       uint64_t tile_num,
       std::vector<Tile>* tiles) const;
 
   /**
    * Generates a new fragment name, which is in the form: <br>
-   * .__uuid_timestamp. For instance,
-   *  __6ba7b8129dad11d180b400c04fd430c8_1458759561320
+   * `__t_t_uuid_v`, where `t` is the input timestamp and `v` is the current
+   * format version. For instance,
+   * `__1458759561320_1458759561320_6ba7b8129dad11d180b400c04fd430c8_3`.
    *
-   * Note that this is a temporary name, initiated by a new write process.
-   * After the new fragmemt is finalized, the array will change its name
-   * by removing the leading '.' character.
+   * If `timestamp` is 0, then it is set to the current time.
    *
+   * @param timestamp The timestamp of when the array got opened for writes. It
+   *     is in ms since 1970-01-01 00:00:00 +0000 (UTC).
    * @param frag_uri Will store the new special fragment name
-   * @param timestamp The timestamp of the fragment name creation.
    * @return Status
    */
-  Status new_fragment_name(std::string* frag_uri, uint64_t* timestamp) const;
+  Status new_fragment_name(
+      uint64_t timestamp, uint32_t format_version, std::string* frag_uri) const;
 
   /**
    * This deletes the global write state and deletes the potentially
@@ -673,6 +497,12 @@ class Writer {
    */
   void optimize_layout_for_1D();
 
+  /**
+   * Checks the validity of the extra element from var-sized offsets of
+   * attributes
+   */
+  Status check_extra_element();
+
   /**
    * Writes in an ordered layout (col- or row-major order). Applicable only
    * to dense arrays.
@@ -688,6 +518,26 @@ class Writer {
   template <class T>
   Status ordered_write();
 
+  /**
+   * Return an element of the offsets buffer at a certain position
+   * taking into account the configured bitsize
+   */
+  uint64_t get_offset_buffer_element(
+      const void* buffer, const uint64_t pos) const;
+
+  /**
+   * Return the size of an offsets buffer according to the configured
+   * options for variable-sized attributes
+   */
+  inline uint64_t get_offset_buffer_size(const uint64_t buffer_size) const;
+
+  /**
+   * Return a buffer offset according to the configured options for
+   * variable-sized attributes (e.g. transform a byte offset to element offset)
+   */
+  uint64_t prepare_buffer_offset(
+      const void* buffer, const uint64_t pos, const uint64_t datasize) const;
+
   /**
    * Applicable only to write in global order. It prepares only full
    * tiles, storing the last potentially non-full tile in
@@ -697,13 +547,30 @@ class Writer {
    * populates the partially full last tile from the previous
    * invocation.
    *
-   * @param attribute The attribute to prepare the tiles for.
    * @param coord_dups The positions of the duplicate coordinates.
    * @param tiles The **full** tiles to be created.
    * @return Status
    */
   Status prepare_full_tiles(
-      const std::string& attribute,
+      const std::set<uint64_t>& coord_dups,
+      std::unordered_map<std::string, std::vector<Tile>>* tiles) const;
+
+  /**
+   * Applicable only to write in global order. It prepares only full
+   * tiles, storing the last potentially non-full tile in
+   * `global_write_state->last_tiles_` as part of the state to be used in
+   * the next write invocation. The last tiles are written to storage
+   * upon `finalize`. Upon each invocation, the function first
+   * populates the partially full last tile from the previous
+   * invocation.
+   *
+   * @param name The attribute/dimension to prepare the tiles for.
+   * @param coord_dups The positions of the duplicate coordinates.
+   * @param tiles The **full** tiles to be created.
+   * @return Status
+   */
+  Status prepare_full_tiles(
+      const std::string& name,
       const std::set<uint64_t>& coord_dups,
       std::vector<Tile>* tiles) const;
 
@@ -716,13 +583,13 @@ class Writer {
    * populates the partially full last tile from the previous
    * invocation. Applicable only to fixed-sized attributes.
    *
-   * @param attribute The attribute to prepare the tiles for.
+   * @param name The attribute/dimension to prepare the tiles for.
    * @param coord_dups The positions of the duplicate coordinates.
    * @param tiles The **full** tiles to be created.
    * @return Status
    */
   Status prepare_full_tiles_fixed(
-      const std::string& attribute,
+      const std::string& name,
       const std::set<uint64_t>& coord_dups,
       std::vector<Tile>* tiles) const;
 
@@ -735,36 +602,39 @@ class Writer {
    * populates the partially full last tile from the previous
    * invocation. Applicable only to var-sized attributes.
    *
-   * @param attribute The attribute to prepare the tiles for.
+   * @param name The attribute/dimension to prepare the tiles for.
    * @param coord_dups The positions of the duplicate coordinates.
    * @param tiles The **full** tiles to be created.
    * @return Status
    */
   Status prepare_full_tiles_var(
-      const std::string& attribute,
+      const std::string& name,
       const std::set<uint64_t>& coord_dups,
       std::vector<Tile>* tiles) const;
 
   /**
-   * It prepares the tiles, copying from the user buffers into the tiles
-   * the values based on the input write cell ranges, focusing on the
-   * input attribute.
+   * It prepares the attribute and coordinate tiles, re-organizing the cells
+   * from the user buffers based on the input sorted positions and coordinate
+   * duplicates.
    *
-   * @param attribute The attribute to prepare the tiles for.
-   * @param write_cell_ranges The write cell ranges.
-   * @param tiles The tiles to be created.
+   * @param cell_pos The positions that resulted from sorting and
+   *     according to which the cells must be re-arranged.
+   * @param coord_dups The set with the positions
+   *     of duplicate coordinates/cells.
+   * @param tiles The tiles to be created, one vector per attribute or
+   *     coordinate.
    * @return Status
    */
   Status prepare_tiles(
-      const std::string& attribute,
-      const std::vector<WriteCellRangeVec>& write_cell_ranges,
-      std::vector<Tile>* tiles) const;
+      const std::vector<uint64_t>& cell_pos,
+      const std::set<uint64_t>& coord_dups,
+      std::unordered_map<std::string, std::vector<Tile>>* tiles) const;
 
   /**
-   * It prepares the tiles, re-organizing the cells from the user
-   * buffers based on the input sorted positions.
+   * It prepares the tiles for the input attribute or dimension, re-organizing
+   * the cells from the user buffers based on the input sorted positions.
    *
-   * @param attribute The attribute to prepare the tiles for.
+   * @param name The attribute or dimension to prepare the tiles for.
    * @param cell_pos The positions that resulted from sorting and
    *     according to which the cells must be re-arranged.
    * @param coord_dups The set with the positions
@@ -773,17 +643,17 @@ class Writer {
    * @return Status
    */
   Status prepare_tiles(
-      const std::string& attribute,
+      const std::string& name,
       const std::vector<uint64_t>& cell_pos,
       const std::set<uint64_t>& coord_dups,
       std::vector<Tile>* tiles) const;
 
   /**
-   * It prepares the tiles, re-organizing the cells from the user
-   * buffers based on the input sorted positions. Applicable only
-   * to fixed-sized attributes.
+   * It prepares the tiles for the input attribute or dimension, re-organizing
+   * the cells from the user buffers based on the input sorted positions.
+   * Applicable only to fixed-sized attributes or dimensions.
    *
-   * @param attribute The attribute to prepare the tiles for.
+   * @param name The attribute or dimension to prepare the tiles for.
    * @param cell_pos The positions that resulted from sorting and
    *     according to which the cells must be re-arranged.
    * @param coord_dups The set with the positions
@@ -792,17 +662,17 @@ class Writer {
    * @return Status
    */
   Status prepare_tiles_fixed(
-      const std::string& attribute,
+      const std::string& name,
       const std::vector<uint64_t>& cell_pos,
       const std::set<uint64_t>& coord_dups,
       std::vector<Tile>* tiles) const;
 
   /**
-   * It prepares the tiles, re-organizing the cells from the user
-   * buffers based on the input sorted positions. Applicable only
-   * to var-sized attributes.
+   * It prepares the tiles for the input attribute or dimension, re-organizing
+   * the cells from the user buffers based on the input sorted positions.
+   * Applicable only to var-sized attributes or dimensions.
    *
-   * @param attribute The attribute to prepare the tiles for.
+   * @param name The attribute to prepare the tiles for.
    * @param cell_pos The positions that resulted from sorting and
    *     according to which the cells must be re-arranged.
    * @param coord_dups The set with the positions
@@ -811,7 +681,7 @@ class Writer {
    * @return Status
    */
   Status prepare_tiles_var(
-      const std::string& attribute,
+      const std::string& name,
       const std::vector<uint64_t>& cell_pos,
       const std::set<uint64_t>& coord_dups,
       std::vector<Tile>* tiles) const;
@@ -823,37 +693,53 @@ class Writer {
    * Sorts the coordinates of the user buffers, creating a vector with
    * the sorted positions.
    *
-   * @tparam T The domain type.
    * @param cell_pos The sorted cell positions to be created.
    * @return Status
    */
-  template <class T>
   Status sort_coords(std::vector<uint64_t>* cell_pos) const;
 
   /**
-   * Writes in unordered layout. Applicable to both dense and sparse arrays.
-   * Explicit coordinates must be provided for this write.
+   * Splits the coordinates buffer into separate coordinate
+   * buffers, one per dimension. Note that this will require extra memory
+   * allocation, which will be cleaned up in the class destructor.
+   *
+   * @return Status
    */
-  Status unordered_write();
+  Status split_coords_buffer();
 
   /**
    * Writes in unordered layout. Applicable to both dense and sparse arrays.
    * Explicit coordinates must be provided for this write.
-   *
-   * @tparam T The domain type.
    */
-  template <class T>
   Status unordered_write();
 
   /**
    * Writes an empty cell range to the input tile.
    * Applicable to **fixed-sized** attributes.
    *
-   * @param num Number of empty values to write.
+   * @param cell_num Number of empty cells to write.
+   * @param cell_val_num Number of values per cell.
    * @param tile The tile to write to.
    * @return Status
    */
-  Status write_empty_cell_range_to_tile(uint64_t num, Tile* tile) const;
+  Status write_empty_cell_range_to_tile(
+      uint64_t num, uint32_t cell_val_num, Tile* tile) const;
+
+  /**
+   * Writes an empty cell range to the input tile.
+   * Applicable to **fixed-sized** attributes.
+   *
+   * @param cell_num Number of empty cells to write.
+   * @param cell_val_num Number of values per cell.
+   * @param tile The tile to write to.
+   * @param tile_validity The tile with the validity cells to write to.
+   * @return Status
+   */
+  Status write_empty_cell_range_to_tile_nullable(
+      uint64_t num,
+      uint32_t cell_val_num,
+      Tile* tile,
+      Tile* tile_validity) const;
 
   /**
    * Writes an empty cell range to the input tile.
@@ -867,6 +753,19 @@ class Writer {
   Status write_empty_cell_range_to_tile_var(
       uint64_t num, Tile* tile, Tile* tile_var) const;
 
+  /**
+   * Writes an empty cell range to the input tile.
+   * Applicable to **variable-sized** attributes.
+   *
+   * @param num Number of empty values to write.
+   * @param tile The tile offsets to write to.
+   * @param tile_var The tile with the var-sized cells to write to.
+   * @param tile_validity The tile with the validity cells to write to.
+   * @return Status
+   */
+  Status write_empty_cell_range_to_tile_var_nullable(
+      uint64_t num, Tile* tile, Tile* tile_var, Tile* tile_validity) const;
+
   /**
    * Writes the input cell range to the input tile, for a particular
    * buffer. Applicable to **fixed-sized** attributes.
@@ -880,6 +779,27 @@ class Writer {
   Status write_cell_range_to_tile(
       ConstBuffer* buff, uint64_t start, uint64_t end, Tile* tile) const;
 
+  /**
+   * Writes the input cell range to the input tile, for a particular
+   * buffer. Applicable to **fixed-sized** attributes.
+   *
+   * @param buff The write buffer where the cells will be copied from.
+   * @param buff_validity The write buffer where the validity cell values will
+   * be copied from.
+   * @param start The start element in the write buffer.
+   * @param end The end element in the write buffer.
+   * @param tile The tile to write to.
+   * @param tile_validity The validity tile to be initialized.
+   * @return Status
+   */
+  Status write_cell_range_to_tile_nullable(
+      ConstBuffer* buff,
+      ConstBuffer* buff_validity,
+      uint64_t start,
+      uint64_t end,
+      Tile* tile,
+      Tile* tile_validity) const;
+
   /**
    * Writes the input cell range to the input tile, for a particular
    * buffer. Applicable to **variable-sized** attributes.
@@ -888,6 +808,7 @@ class Writer {
    * @param buff_var The write buffer where the cell values will be copied from.
    * @param start The start element in the write buffer.
    * @param end The end element in the write buffer.
+   * @param attr_datatype_size The size of each attribute value in `buff_var`.
    * @param tile The tile offsets to write to.
    * @param tile_var The tile with the var-sized cells to write to.
    * @return Status
@@ -897,31 +818,107 @@ class Writer {
       ConstBuffer* buff_var,
       uint64_t start,
       uint64_t end,
+      uint64_t attr_datatype_size,
       Tile* tile,
       Tile* tile_var) const;
 
+  /**
+   * Writes the input cell range to the input tile, for a particular
+   * buffer. Applicable to **variable-sized**, nullable attributes.
+   *
+   * @param buff The write buffer where the cell offsets will be copied from.
+   * @param buff_var The write buffer where the cell values will be copied from.
+   * @param buff_validity The write buffer where the validity cell values will
+   * be copied from.
+   * @param start The start element in the write buffer.
+   * @param end The end element in the write buffer.
+   * @param attr_datatype_size The size of each attribute value in `buff_var`.
+   * @param tile The tile offsets to write to.
+   * @param tile_var The tile with the var-sized cells to write to.
+   * @param tile_validity The validity tile to be initialized.
+   * @return Status
+   */
+  Status write_cell_range_to_tile_var_nullable(
+      ConstBuffer* buff,
+      ConstBuffer* buff_var,
+      ConstBuffer* buff_validity,
+      uint64_t start,
+      uint64_t end,
+      uint64_t attr_datatype_size,
+      Tile* tile,
+      Tile* tile_var,
+      Tile* tile_validity) const;
+
   /**
    * Writes all the input tiles to storage.
    *
-   * @param attribute_tiles Tiles to be written, one element per attribute.
+   * @param tiles Attribute/Coordinate tiles to be written, one element per
+   *     attribute or dimension.
+   * @param tiles Attribute/Coordinate tiles to be written.
    * @return Status
    */
   Status write_all_tiles(
-      FragmentMetadata* frag_meta,
-      const std::vector<std::vector<Tile>>& attribute_tiles) const;
+      tdb_shared_ptr<FragmentMetadata> frag_meta,
+      std::unordered_map<std::string, std::vector<Tile>>* tiles);
 
   /**
-   * Writes the input tiles for the input attribute to storage.
+   * Writes the input tiles for the input attribute/dimension to storage.
    *
-   * @param attribute The attribute the tiles belong to.
+   * @param name The attribute/dimension the tiles belong to.
    * @param frag_meta The fragment metadata.
+   * @param start_tile_id The function will start writing tiles
+   *     with ids in the fragment that start with this value.
    * @param tiles The tiles to be written.
+   * @param close_files Whether to close the attribute/coordinate
+   *     file in the end of the function call.
    * @return Status
    */
   Status write_tiles(
-      const std::string& attribute,
-      FragmentMetadata* frag_meta,
-      const std::vector<Tile>& tiles) const;
+      const std::string& name,
+      tdb_shared_ptr<FragmentMetadata> frag_meta,
+      uint64_t start_tile_id,
+      std::vector<Tile>* tiles,
+      bool close_files = true);
+
+  /**
+   * Returns the i-th coordinates in the coordinate buffers in string
+   * format.
+   */
+  std::string coords_to_str(uint64_t i) const;
+
+  /**
+   * Invoked on error. It removes the directory of the input URI and
+   * resets the global write state.
+   */
+  void clean_up(const URI& uri);
+
+  /**
+   * Applicable only to global writes. Returns true if all last tiles stored
+   * in the global write state are empty.
+   */
+  bool all_last_tiles_empty() const;
+
+  /** Calculates the hilbert values of the input coordinate buffers. */
+  Status calculate_hilbert_values(
+      const std::vector<const QueryBuffer*>& buffs,
+      std::vector<uint64_t>* hilbert_values) const;
+
+  /**
+   * Prepares, filters and writes dense tiles for the given attribute.
+   *
+   * @tparam T The array domain datatype.
+   * @param name The attribute name.
+   * @param frag_meta The metadata of the new fragment.
+   * @param dense_tiler The dense tiler that will prepare the tiles.
+   * @param thread_num The number of threads to be used for the function.
+   * @param stats Statistics to gather in the function.
+   */
+  template <class T>
+  Status prepare_filter_and_write_tiles(
+      const std::string& name,
+      tdb_shared_ptr<FragmentMetadata> frag_meta,
+      DenseTiler<T>* dense_tiler,
+      uint64_t thread_num);
 };
 
 }  // namespace sm
