@@ -5,7 +5,7 @@
  *
  * The MIT License
  *
- * @copyright Copyright (c) 2018-2019 TileDB, Inc.
+ * @copyright Copyright (c) 2018-2021 TileDB, Inc.
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
  * of this software and associated documentation files (the "Software"), to deal
@@ -29,23 +29,32 @@
  *
  * This file defines serialization functions for ArraySchema.
  */
-
-#include "tiledb/sm/serialization/array_schema.h"
-#include "tiledb/sm/enums/array_type.h"
-#include "tiledb/sm/enums/compressor.h"
-#include "tiledb/sm/enums/datatype.h"
-#include "tiledb/sm/enums/layout.h"
-#include "tiledb/sm/enums/serialization_type.h"
-#include "tiledb/sm/misc/constants.h"
-#include "tiledb/sm/misc/logger.h"
-#include "tiledb/sm/misc/stats.h"
-#include "tiledb/sm/serialization/capnp_utils.h"
-
 #ifdef TILEDB_SERIALIZATION
 #include <capnp/compat/json.h>
 #include <capnp/serialize.h>
+#include "tiledb/sm/serialization/capnp_utils.h"
 #endif
 
+#include "tiledb/common/heap_memory.h"
+#include "tiledb/common/logger.h"
+#include "tiledb/sm/array/array.h"
+#include "tiledb/sm/array_schema/attribute.h"
+#include "tiledb/sm/array_schema/dimension.h"
+#include "tiledb/sm/array_schema/domain.h"
+#include "tiledb/sm/enums/array_type.h"
+#include "tiledb/sm/enums/compressor.h"
+#include "tiledb/sm/enums/datatype.h"
+#include "tiledb/sm/enums/filter_option.h"
+#include "tiledb/sm/enums/filter_type.h"
+#include "tiledb/sm/enums/layout.h"
+#include "tiledb/sm/enums/serialization_type.h"
+#include "tiledb/sm/misc/constants.h"
+#include "tiledb/sm/serialization/array_schema.h"
+
+#include <set>
+
+using namespace tiledb::common;
+
 namespace tiledb {
 namespace sm {
 namespace serialization {
@@ -109,16 +118,16 @@ Status filter_pipeline_to_capnp(
 
 Status filter_pipeline_from_capnp(
     const capnp::FilterPipeline::Reader& filter_pipeline_reader,
-    std::unique_ptr<FilterPipeline>* filter_pipeline) {
-  filter_pipeline->reset(new FilterPipeline);
+    tdb_unique_ptr<FilterPipeline>* filter_pipeline) {
+  filter_pipeline->reset(tdb_new(FilterPipeline));
   if (!filter_pipeline_reader.hasFilters())
     return Status::Ok();
 
   auto filter_list_reader = filter_pipeline_reader.getFilters();
-  for (const auto& filter_reader : filter_list_reader) {
+  for (auto filter_reader : filter_list_reader) {
     FilterType type = FilterType::FILTER_NONE;
     RETURN_NOT_OK(filter_type_enum(filter_reader.getType().cStr(), &type));
-    std::unique_ptr<Filter> filter(Filter::create(type));
+    tdb_unique_ptr<Filter> filter(Filter::create(type));
     if (filter == nullptr)
       return LOG_STATUS(Status::SerializationError(
           "Error deserializing filter pipeline; failed to create filter."));
@@ -169,32 +178,77 @@ Status attribute_to_capnp(
   attribute_builder->setName(attribute->name());
   attribute_builder->setType(datatype_str(attribute->type()));
   attribute_builder->setCellValNum(attribute->cell_val_num());
+  attribute_builder->setNullable(attribute->nullable());
 
-  const auto* filters = attribute->filters();
+  // Get the fill value from `attribute`.
+  const void* fill_value;
+  uint64_t fill_value_size;
+  uint8_t fill_validity = true;
+  if (!attribute->nullable())
+    RETURN_NOT_OK(attribute->get_fill_value(&fill_value, &fill_value_size));
+  else
+    RETURN_NOT_OK(attribute->get_fill_value(
+        &fill_value, &fill_value_size, &fill_validity));
+
+  // Copy the fill value buffer into a capnp vector of bytes.
+  auto capnpFillValue = kj::Vector<uint8_t>();
+  capnpFillValue.addAll(kj::ArrayPtr<uint8_t>(
+      const_cast<uint8_t*>(reinterpret_cast<const uint8_t*>(fill_value)),
+      fill_value_size));
+
+  // Store the fill value vector of bytes.
+  attribute_builder->setFillValue(capnpFillValue.asPtr());
+
+  // Set the validity fill value.
+  attribute_builder->setFillValueValidity(fill_validity ? true : false);
+
+  const auto& filters = attribute->filters();
   auto filter_pipeline_builder = attribute_builder->initFilterPipeline();
-  RETURN_NOT_OK(filter_pipeline_to_capnp(filters, &filter_pipeline_builder));
+  RETURN_NOT_OK(filter_pipeline_to_capnp(&filters, &filter_pipeline_builder));
 
   return Status::Ok();
 }
 
 Status attribute_from_capnp(
     const capnp::Attribute::Reader& attribute_reader,
-    std::unique_ptr<Attribute>* attribute) {
+    tdb_unique_ptr<Attribute>* attribute) {
   Datatype datatype = Datatype::ANY;
   RETURN_NOT_OK(datatype_enum(attribute_reader.getType(), &datatype));
 
-  attribute->reset(new Attribute(attribute_reader.getName(), datatype));
+  attribute->reset(tdb_new(Attribute, attribute_reader.getName(), datatype));
   RETURN_NOT_OK(
       (*attribute)->set_cell_val_num(attribute_reader.getCellValNum()));
 
-  // Set filter pipelines
+  // Set nullable.
+  const bool nullable = attribute_reader.getNullable();
+  RETURN_NOT_OK((*attribute)->set_nullable(nullable));
+
+  // Set the fill value.
+  if (attribute_reader.hasFillValue()) {
+    auto fill_value = attribute_reader.getFillValue();
+    if (nullable) {
+      (*attribute)
+          ->set_fill_value(
+              fill_value.asBytes().begin(),
+              fill_value.size(),
+              attribute_reader.getFillValueValidity());
+    } else {
+      (*attribute)
+          ->set_fill_value(fill_value.asBytes().begin(), fill_value.size());
+    }
+  }
+
+  // Set filter pipelines.
   if (attribute_reader.hasFilterPipeline()) {
     auto filter_pipeline_reader = attribute_reader.getFilterPipeline();
-    std::unique_ptr<FilterPipeline> filters;
+    tdb_unique_ptr<FilterPipeline> filters;
     RETURN_NOT_OK(filter_pipeline_from_capnp(filter_pipeline_reader, &filters));
     RETURN_NOT_OK((*attribute)->set_filter_pipeline(filters.get()));
   }
 
+  // Set nullable.
+  RETURN_NOT_OK((*attribute)->set_nullable(attribute_reader.getNullable()));
+
   return Status::Ok();
 }
 
@@ -206,33 +260,54 @@ Status dimension_to_capnp(
 
   dimension_builder->setName(dimension->name());
   dimension_builder->setType(datatype_str(dimension->type()));
-  dimension_builder->setNullTileExtent(dimension->tile_extent() == nullptr);
+  dimension_builder->setNullTileExtent(!dimension->tile_extent());
 
-  auto domain_builder = dimension_builder->initDomain();
-  RETURN_NOT_OK(utils::set_capnp_array_ptr(
-      domain_builder, dimension->type(), dimension->domain(), 2));
-
-  if (dimension->tile_extent() != nullptr) {
-    auto tile_extent_builder = dimension_builder->initTileExtent();
-    RETURN_NOT_OK(utils::set_capnp_scalar(
-        tile_extent_builder, dimension->type(), dimension->tile_extent()));
+  // Only set the domain if its not empty/null. String dimensions have null
+  // domains
+  if (!dimension->domain().empty()) {
+    auto domain_builder = dimension_builder->initDomain();
+    RETURN_NOT_OK(utils::set_capnp_array_ptr(
+        domain_builder, dimension->type(), dimension->domain().data(), 2));
   }
 
+  // Only set the tile extent if its not empty
+  if (dimension->tile_extent()) {
+    auto tile_extent_builder = dimension_builder->initTileExtent();
+    RETURN_NOT_OK(utils::set_capnp_scalar(
+        tile_extent_builder,
+        dimension->type(),
+        dimension->tile_extent().data()));
+  }
+
+  // Set filters
+  const FilterPipeline& coords_filters = dimension->filters();
+  capnp::FilterPipeline::Builder filters_builder =
+      dimension_builder->initFilterPipeline();
+  RETURN_NOT_OK(filter_pipeline_to_capnp(&coords_filters, &filters_builder));
   return Status::Ok();
 }
 
 Status dimension_from_capnp(
     const capnp::Dimension::Reader& dimension_reader,
-    std::unique_ptr<Dimension>* dimension) {
+    tdb_unique_ptr<Dimension>* dimension) {
   Datatype dim_type = Datatype::ANY;
   RETURN_NOT_OK(datatype_enum(dimension_reader.getType().cStr(), &dim_type));
-  dimension->reset(new Dimension(dimension_reader.getName(), dim_type));
+  dimension->reset(tdb_new(Dimension, dimension_reader.getName(), dim_type));
 
-  auto domain_reader = dimension_reader.getDomain();
-  Buffer domain_buffer;
-  RETURN_NOT_OK(
-      utils::copy_capnp_list(domain_reader, dim_type, &domain_buffer));
-  RETURN_NOT_OK((*dimension)->set_domain(domain_buffer.data()));
+  if (dimension_reader.hasDomain()) {
+    auto domain_reader = dimension_reader.getDomain();
+    Buffer domain_buffer;
+    RETURN_NOT_OK(
+        utils::copy_capnp_list(domain_reader, dim_type, &domain_buffer));
+    RETURN_NOT_OK((*dimension)->set_domain_unsafe(domain_buffer.data()));
+  }
+
+  if (dimension_reader.hasFilterPipeline()) {
+    auto reader = dimension_reader.getFilterPipeline();
+    tdb_unique_ptr<FilterPipeline> filters;
+    RETURN_NOT_OK(filter_pipeline_from_capnp(reader, &filters));
+    RETURN_NOT_OK((*dimension)->set_filter_pipeline(filters.get()));
+  }
 
   if (!dimension_reader.getNullTileExtent()) {
     auto tile_extent_reader = dimension_reader.getTileExtent();
@@ -280,6 +355,15 @@ Status dimension_from_capnp(
       case Datatype::DATETIME_PS:
       case Datatype::DATETIME_FS:
       case Datatype::DATETIME_AS:
+      case Datatype::TIME_HR:
+      case Datatype::TIME_MIN:
+      case Datatype::TIME_SEC:
+      case Datatype::TIME_MS:
+      case Datatype::TIME_US:
+      case Datatype::TIME_NS:
+      case Datatype::TIME_PS:
+      case Datatype::TIME_FS:
+      case Datatype::TIME_AS:
       case Datatype::INT64: {
         auto val = tile_extent_reader.getInt64();
         RETURN_NOT_OK((*dimension)->set_tile_extent(&val));
@@ -315,7 +399,7 @@ Status domain_to_capnp(
     return LOG_STATUS(Status::SerializationError(
         "Error serializing domain; domain is null."));
 
-  domainBuilder->setType(datatype_str(domain->type()));
+  domainBuilder->setType(datatype_str(domain->dimension(0)->type()));
   domainBuilder->setTileOrder(layout_str(domain->tile_order()));
   domainBuilder->setCellOrder(layout_str(domain->cell_order()));
 
@@ -331,14 +415,14 @@ Status domain_to_capnp(
 
 Status domain_from_capnp(
     const capnp::Domain::Reader& domain_reader,
-    std::unique_ptr<Domain>* domain) {
+    tdb_unique_ptr<Domain>* domain) {
   Datatype datatype = Datatype::ANY;
   RETURN_NOT_OK(datatype_enum(domain_reader.getType(), &datatype));
-  domain->reset(new Domain(datatype));
+  domain->reset(tdb_new(Domain));
 
   auto dimensions = domain_reader.getDimensions();
-  for (const auto& dimension : dimensions) {
-    std::unique_ptr<Dimension> dim;
+  for (auto dimension : dimensions) {
+    tdb_unique_ptr<Dimension> dim;
     RETURN_NOT_OK(dimension_from_capnp(dimension, &dim));
     RETURN_NOT_OK((*domain)->add_dimension(dim.get()));
   }
@@ -348,33 +432,47 @@ Status domain_from_capnp(
 
 Status array_schema_to_capnp(
     const ArraySchema* array_schema,
-    capnp::ArraySchema::Builder* array_schema_builder) {
+    capnp::ArraySchema::Builder* array_schema_builder,
+    const bool client_side) {
   if (array_schema == nullptr)
     return LOG_STATUS(Status::SerializationError(
         "Error serializing array schema; array schema is null."));
 
-  array_schema_builder->setUri(array_schema->array_uri().to_string());
-  array_schema_builder->setVersion(kj::arrayPtr(constants::library_version, 3));
+  // Only set the URI if client side
+  if (client_side)
+    array_schema_builder->setUri(array_schema->array_uri().to_string());
+  auto v = kj::heapArray<int32_t>(1);
+  v[0] = array_schema->version();
+  array_schema_builder->setVersion(v);
   array_schema_builder->setArrayType(
       array_type_str(array_schema->array_type()));
   array_schema_builder->setTileOrder(layout_str(array_schema->tile_order()));
   array_schema_builder->setCellOrder(layout_str(array_schema->cell_order()));
   array_schema_builder->setCapacity(array_schema->capacity());
+  array_schema_builder->setAllowsDuplicates(array_schema->allows_dups());
 
   // Set coordinate filters
-  const FilterPipeline* coords_filters = array_schema->coords_filters();
+  const FilterPipeline& coords_filters = array_schema->coords_filters();
   capnp::FilterPipeline::Builder coords_filters_builder =
       array_schema_builder->initCoordsFilterPipeline();
   RETURN_NOT_OK(
-      filter_pipeline_to_capnp(coords_filters, &coords_filters_builder));
+      filter_pipeline_to_capnp(&coords_filters, &coords_filters_builder));
 
   // Set offset filters
-  const FilterPipeline* offsets_filters =
+  const FilterPipeline& offsets_filters =
       array_schema->cell_var_offsets_filters();
   capnp::FilterPipeline::Builder offsets_filters_builder =
       array_schema_builder->initOffsetFilterPipeline();
   RETURN_NOT_OK(
-      filter_pipeline_to_capnp(offsets_filters, &offsets_filters_builder));
+      filter_pipeline_to_capnp(&offsets_filters, &offsets_filters_builder));
+
+  // Set validity filters
+  const FilterPipeline& validity_filters =
+      array_schema->cell_validity_filters();
+  capnp::FilterPipeline::Builder validity_filters_builder =
+      array_schema_builder->initValidityFilterPipeline();
+  RETURN_NOT_OK(
+      filter_pipeline_to_capnp(&validity_filters, &validity_filters_builder));
 
   // Domain
   auto domain_builder = array_schema_builder->initDomain();
@@ -394,29 +492,38 @@ Status array_schema_to_capnp(
 
 Status array_schema_from_capnp(
     const capnp::ArraySchema::Reader& schema_reader,
-    std::unique_ptr<ArraySchema>* array_schema) {
+    tdb_unique_ptr<ArraySchema>* array_schema) {
   ArrayType array_type = ArrayType::DENSE;
   RETURN_NOT_OK(array_type_enum(schema_reader.getArrayType(), &array_type));
-  array_schema->reset(new ArraySchema(array_type));
+  array_schema->reset(tdb_new(ArraySchema, array_type));
 
   Layout layout = Layout::ROW_MAJOR;
   RETURN_NOT_OK(layout_enum(schema_reader.getTileOrder().cStr(), &layout));
   (*array_schema)->set_tile_order(layout);
   RETURN_NOT_OK(layout_enum(schema_reader.getCellOrder().cStr(), &layout));
 
-  (*array_schema)->set_array_uri(URI(schema_reader.getUri().cStr()));
+  if (schema_reader.hasUri())
+    (*array_schema)->set_array_uri(URI(schema_reader.getUri().cStr()));
+
   (*array_schema)->set_cell_order(layout);
   (*array_schema)->set_capacity(schema_reader.getCapacity());
+  (*array_schema)->set_allows_dups(schema_reader.getAllowsDuplicates());
+  // Pre 1.8 TileDB serialized the version as the library version
+  // This would have been a list of size 3, so only set the version
+  // if the list size is 1, meaning tiledb 1.8 or later
+  if (schema_reader.hasVersion() && schema_reader.getVersion().size() == 1) {
+    (*array_schema)->set_version(schema_reader.getVersion()[0]);
+  }
 
   auto domain_reader = schema_reader.getDomain();
-  std::unique_ptr<Domain> domain;
+  tdb_unique_ptr<Domain> domain;
   RETURN_NOT_OK(domain_from_capnp(domain_reader, &domain));
   RETURN_NOT_OK((*array_schema)->set_domain(domain.get()));
 
   // Set coords filter pipelines
   if (schema_reader.hasCoordsFilterPipeline()) {
     auto reader = schema_reader.getCoordsFilterPipeline();
-    std::unique_ptr<FilterPipeline> filters;
+    tdb_unique_ptr<FilterPipeline> filters;
     RETURN_NOT_OK(filter_pipeline_from_capnp(reader, &filters));
     RETURN_NOT_OK((*array_schema)->set_coords_filter_pipeline(filters.get()));
   }
@@ -424,16 +531,25 @@ Status array_schema_from_capnp(
   // Set offsets filter pipelines
   if (schema_reader.hasOffsetFilterPipeline()) {
     auto reader = schema_reader.getOffsetFilterPipeline();
-    std::unique_ptr<FilterPipeline> filters;
+    tdb_unique_ptr<FilterPipeline> filters;
     RETURN_NOT_OK(filter_pipeline_from_capnp(reader, &filters));
     RETURN_NOT_OK(
         (*array_schema)->set_cell_var_offsets_filter_pipeline(filters.get()));
   }
 
+  // Set validity filter pipelines
+  if (schema_reader.hasValidityFilterPipeline()) {
+    auto reader = schema_reader.getValidityFilterPipeline();
+    tdb_unique_ptr<FilterPipeline> filters;
+    RETURN_NOT_OK(filter_pipeline_from_capnp(reader, &filters));
+    RETURN_NOT_OK(
+        (*array_schema)->set_cell_validity_filter_pipeline(filters.get()));
+  }
+
   // Set attributes
   auto attributes_reader = schema_reader.getAttributes();
-  for (const auto& attr_reader : attributes_reader) {
-    std::unique_ptr<Attribute> attribute;
+  for (auto attr_reader : attributes_reader) {
+    tdb_unique_ptr<Attribute> attribute;
     RETURN_NOT_OK(attribute_from_capnp(attr_reader, &attribute));
     RETURN_NOT_OK((*array_schema)->add_attribute(attribute.get(), false));
   }
@@ -447,14 +563,14 @@ Status array_schema_from_capnp(
 Status array_schema_serialize(
     ArraySchema* array_schema,
     SerializationType serialize_type,
-    Buffer* serialized_buffer) {
-  STATS_FUNC_IN(serialization_array_schema_serialize);
-
+    Buffer* serialized_buffer,
+    const bool client_side) {
   try {
     ::capnp::MallocMessageBuilder message;
     capnp::ArraySchema::Builder arraySchemaBuilder =
         message.initRoot<capnp::ArraySchema>();
-    RETURN_NOT_OK(array_schema_to_capnp(array_schema, &arraySchemaBuilder));
+    RETURN_NOT_OK(
+        array_schema_to_capnp(array_schema, &arraySchemaBuilder, client_side));
 
     serialized_buffer->reset_size();
     serialized_buffer->reset_offset();
@@ -496,18 +612,14 @@ Status array_schema_serialize(
   }
 
   return Status::Ok();
-
-  STATS_FUNC_OUT(serialization_array_schema_serialize);
 }
 
 Status array_schema_deserialize(
     ArraySchema** array_schema,
     SerializationType serialize_type,
     const Buffer& serialized_buffer) {
-  STATS_FUNC_IN(serialization_array_schema_deserialize);
-
   try {
-    std::unique_ptr<ArraySchema> decoded_array_schema = nullptr;
+    tdb_unique_ptr<ArraySchema> decoded_array_schema = nullptr;
 
     switch (serialize_type) {
       case SerializationType::JSON: {
@@ -559,8 +671,144 @@ Status array_schema_deserialize(
   }
 
   return Status::Ok();
+}
 
-  STATS_FUNC_OUT(serialization_array_schema_deserialize);
+Status nonempty_domain_serialize(
+    const Dimension* dimension,
+    const void* nonempty_domain,
+    bool is_empty,
+    SerializationType serialize_type,
+    Buffer* serialized_buffer) {
+  if (!is_empty && nonempty_domain == nullptr)
+    return LOG_STATUS(Status::SerializationError(
+        "Error serializing nonempty domain; nonempty domain is null."));
+
+  try {
+    // Serialize
+    ::capnp::MallocMessageBuilder message;
+    auto builder = message.initRoot<capnp::NonEmptyDomain>();
+    builder.setIsEmpty(is_empty);
+
+    if (!is_empty) {
+      auto subarray_builder = builder.initNonEmptyDomain();
+      RETURN_NOT_OK(utils::serialize_coords(
+          subarray_builder, dimension, nonempty_domain));
+    }
+
+    // Copy to buffer
+    serialized_buffer->reset_size();
+    serialized_buffer->reset_offset();
+
+    switch (serialize_type) {
+      case SerializationType::JSON: {
+        ::capnp::JsonCodec json;
+        kj::String capnp_json = json.encode(builder);
+        const auto json_len = capnp_json.size();
+        const char nul = '\0';
+        // size does not include needed null terminator, so add +1
+        RETURN_NOT_OK(serialized_buffer->realloc(json_len + 1));
+        RETURN_NOT_OK(serialized_buffer->write(capnp_json.cStr(), json_len));
+        RETURN_NOT_OK(serialized_buffer->write(&nul, 1));
+        break;
+      }
+      case SerializationType::CAPNP: {
+        kj::Array<::capnp::word> protomessage = messageToFlatArray(message);
+        kj::ArrayPtr<const char> message_chars = protomessage.asChars();
+        const auto nbytes = message_chars.size();
+        RETURN_NOT_OK(serialized_buffer->realloc(nbytes));
+        RETURN_NOT_OK(serialized_buffer->write(message_chars.begin(), nbytes));
+        break;
+      }
+      default: {
+        return LOG_STATUS(Status::SerializationError(
+            "Error serializing nonempty domain; Unknown serialization type "
+            "passed"));
+      }
+    }
+
+  } catch (kj::Exception& e) {
+    return LOG_STATUS(Status::SerializationError(
+        "Error serializing nonempty domain; kj::Exception: " +
+        std::string(e.getDescription().cStr())));
+  } catch (std::exception& e) {
+    return LOG_STATUS(Status::SerializationError(
+        "Error serializing nonempty domain; exception " +
+        std::string(e.what())));
+  }
+
+  return Status::Ok();
+}
+
+Status nonempty_domain_deserialize(
+    const Dimension* dimension,
+    const Buffer& serialized_buffer,
+    SerializationType serialize_type,
+    void* nonempty_domain,
+    bool* is_empty) {
+  if (nonempty_domain == nullptr)
+    return LOG_STATUS(Status::SerializationError(
+        "Error deserializing nonempty domain; nonempty domain is null."));
+
+  try {
+    switch (serialize_type) {
+      case SerializationType::JSON: {
+        ::capnp::JsonCodec json;
+        ::capnp::MallocMessageBuilder message_builder;
+        auto builder = message_builder.initRoot<capnp::NonEmptyDomain>();
+        json.decode(
+            kj::StringPtr(static_cast<const char*>(serialized_buffer.data())),
+            builder);
+        auto reader = builder.asReader();
+
+        // Deserialize
+        *is_empty = reader.getIsEmpty();
+        if (!*is_empty) {
+          void* subarray;
+          RETURN_NOT_OK(utils::deserialize_coords(
+              reader.getNonEmptyDomain(), dimension, &subarray));
+          std::memcpy(nonempty_domain, subarray, 2 * dimension->coord_size());
+          tdb_free(subarray);
+        }
+
+        break;
+      }
+      case SerializationType::CAPNP: {
+        const auto mBytes =
+            reinterpret_cast<const kj::byte*>(serialized_buffer.data());
+        ::capnp::FlatArrayMessageReader msg_reader(kj::arrayPtr(
+            reinterpret_cast<const ::capnp::word*>(mBytes),
+            serialized_buffer.size() / sizeof(::capnp::word)));
+        auto reader = msg_reader.getRoot<capnp::NonEmptyDomain>();
+
+        // Deserialize
+        *is_empty = reader.getIsEmpty();
+        if (!*is_empty) {
+          void* subarray;
+          RETURN_NOT_OK(utils::deserialize_coords(
+              reader.getNonEmptyDomain(), dimension, &subarray));
+          std::memcpy(nonempty_domain, subarray, 2 * dimension->coord_size());
+          tdb_free(subarray);
+        }
+
+        break;
+      }
+      default: {
+        return LOG_STATUS(Status::SerializationError(
+            "Error deserializing nonempty domain; Unknown serialization type "
+            "passed"));
+      }
+    }
+  } catch (kj::Exception& e) {
+    return LOG_STATUS(Status::SerializationError(
+        "Error deserializing nonempty domain; kj::Exception: " +
+        std::string(e.getDescription().cStr())));
+  } catch (std::exception& e) {
+    return LOG_STATUS(Status::SerializationError(
+        "Error deserializing nonempty domain; exception " +
+        std::string(e.what())));
+  }
+
+  return Status::Ok();
 }
 
 Status nonempty_domain_serialize(
@@ -573,7 +821,7 @@ Status nonempty_domain_serialize(
     return LOG_STATUS(Status::SerializationError(
         "Error serializing nonempty domain; nonempty domain is null."));
 
-  const auto* schema = array->array_schema();
+  const auto* schema = array->array_schema_latest();
   if (schema == nullptr)
     return LOG_STATUS(Status::SerializationError(
         "Error serializing nonempty domain; array schema is null."));
@@ -644,7 +892,7 @@ Status nonempty_domain_deserialize(
     return LOG_STATUS(Status::SerializationError(
         "Error deserializing nonempty domain; nonempty domain is null."));
 
-  const auto* schema = array->array_schema();
+  const auto* schema = array->array_schema_latest();
   if (schema == nullptr)
     return LOG_STATUS(Status::SerializationError(
         "Error deserializing nonempty domain; array schema is null."));
@@ -666,8 +914,11 @@ Status nonempty_domain_deserialize(
           void* subarray;
           RETURN_NOT_OK(utils::deserialize_subarray(
               reader.getNonEmptyDomain(), schema, &subarray));
-          std::memcpy(nonempty_domain, subarray, 2 * schema->coords_size());
-          std::free(subarray);
+          std::memcpy(
+              nonempty_domain,
+              subarray,
+              2 * schema->dimension(0)->coord_size());
+          tdb_free(subarray);
         }
 
         break;
@@ -686,8 +937,11 @@ Status nonempty_domain_deserialize(
           void* subarray;
           RETURN_NOT_OK(utils::deserialize_subarray(
               reader.getNonEmptyDomain(), schema, &subarray));
-          std::memcpy(nonempty_domain, subarray, 2 * schema->coords_size());
-          std::free(subarray);
+          std::memcpy(
+              nonempty_domain,
+              subarray,
+              2 * schema->dimension(0)->coord_size());
+          tdb_free(subarray);
         }
 
         break;
@@ -711,12 +965,120 @@ Status nonempty_domain_deserialize(
   return Status::Ok();
 }
 
+Status nonempty_domain_serialize(
+    Array* array, SerializationType serialize_type, Buffer* serialized_buffer) {
+  const auto* schema = array->array_schema_latest();
+  if (schema == nullptr)
+    return LOG_STATUS(Status::SerializationError(
+        "Error serializing nonempty domain; array schema is null."));
+
+  try {
+    // Serialize
+    ::capnp::MallocMessageBuilder message;
+    auto builder = message.initRoot<capnp::NonEmptyDomainList>();
+
+    RETURN_NOT_OK(utils::serialize_non_empty_domain(builder, array));
+
+    // Copy to buffer
+    serialized_buffer->reset_size();
+    serialized_buffer->reset_offset();
+
+    switch (serialize_type) {
+      case SerializationType::JSON: {
+        ::capnp::JsonCodec json;
+        kj::String capnp_json = json.encode(builder);
+        const auto json_len = capnp_json.size();
+        const char nul = '\0';
+        // size does not include needed null terminator, so add +1
+        RETURN_NOT_OK(serialized_buffer->realloc(json_len + 1));
+        RETURN_NOT_OK(serialized_buffer->write(capnp_json.cStr(), json_len));
+        RETURN_NOT_OK(serialized_buffer->write(&nul, 1));
+        break;
+      }
+      case SerializationType::CAPNP: {
+        kj::Array<::capnp::word> protomessage = messageToFlatArray(message);
+        kj::ArrayPtr<const char> message_chars = protomessage.asChars();
+        const auto nbytes = message_chars.size();
+        RETURN_NOT_OK(serialized_buffer->realloc(nbytes));
+        RETURN_NOT_OK(serialized_buffer->write(message_chars.begin(), nbytes));
+        break;
+      }
+      default: {
+        return LOG_STATUS(Status::SerializationError(
+            "Error serializing nonempty domain; Unknown serialization type "
+            "passed"));
+      }
+    }
+
+  } catch (kj::Exception& e) {
+    return LOG_STATUS(Status::SerializationError(
+        "Error serializing nonempty domain; kj::Exception: " +
+        std::string(e.getDescription().cStr())));
+  } catch (std::exception& e) {
+    return LOG_STATUS(Status::SerializationError(
+        "Error serializing nonempty domain; exception " +
+        std::string(e.what())));
+  }
+
+  return Status::Ok();
+}
+
+Status nonempty_domain_deserialize(
+    Array* array,
+    const Buffer& serialized_buffer,
+    SerializationType serialize_type) {
+  try {
+    switch (serialize_type) {
+      case SerializationType::JSON: {
+        ::capnp::JsonCodec json;
+        ::capnp::MallocMessageBuilder message_builder;
+        auto builder = message_builder.initRoot<capnp::NonEmptyDomainList>();
+        json.decode(
+            kj::StringPtr(static_cast<const char*>(serialized_buffer.data())),
+            builder);
+        auto reader = builder.asReader();
+
+        // Deserialize
+        RETURN_NOT_OK(utils::deserialize_non_empty_domain(reader, array));
+        break;
+      }
+      case SerializationType::CAPNP: {
+        const auto mBytes =
+            reinterpret_cast<const kj::byte*>(serialized_buffer.data());
+        ::capnp::FlatArrayMessageReader msg_reader(kj::arrayPtr(
+            reinterpret_cast<const ::capnp::word*>(mBytes),
+            serialized_buffer.size() / sizeof(::capnp::word)));
+        auto reader = msg_reader.getRoot<capnp::NonEmptyDomainList>();
+
+        // Deserialize
+        RETURN_NOT_OK(utils::deserialize_non_empty_domain(reader, array));
+        break;
+      }
+      default: {
+        return LOG_STATUS(Status::SerializationError(
+            "Error deserializing nonempty domain; Unknown serialization type "
+            "passed"));
+      }
+    }
+  } catch (kj::Exception& e) {
+    return LOG_STATUS(Status::SerializationError(
+        "Error deserializing nonempty domain; kj::Exception: " +
+        std::string(e.getDescription().cStr())));
+  } catch (std::exception& e) {
+    return LOG_STATUS(Status::SerializationError(
+        "Error deserializing nonempty domain; exception " +
+        std::string(e.what())));
+  }
+
+  return Status::Ok();
+}
+
 Status max_buffer_sizes_serialize(
     Array* array,
     const void* subarray,
     SerializationType serialize_type,
     Buffer* serialized_buffer) {
-  const auto* schema = array->array_schema();
+  const auto* schema = array->array_schema_latest();
   if (schema == nullptr)
     return LOG_STATUS(Status::SerializationError(
         "Error serializing max buffer sizes; array schema is null."));
@@ -1069,7 +1431,8 @@ Status array_metadata_deserialize(
 
 #else
 
-Status array_schema_serialize(ArraySchema*, SerializationType, Buffer*) {
+Status array_schema_serialize(
+    ArraySchema*, SerializationType, Buffer*, const bool) {
   return LOG_STATUS(Status::SerializationError(
       "Cannot serialize; serialization not enabled."));
 }
@@ -1080,6 +1443,16 @@ Status array_schema_deserialize(
       "Cannot serialize; serialization not enabled."));
 }
 
+Status nonempty_domain_serialize(Array*, SerializationType, Buffer*) {
+  return LOG_STATUS(Status::SerializationError(
+      "Cannot serialize; serialization not enabled."));
+}
+
+Status nonempty_domain_deserialize(Array*, const Buffer&, SerializationType) {
+  return LOG_STATUS(Status::SerializationError(
+      "Cannot serialize; serialization not enabled."));
+}
+
 Status nonempty_domain_serialize(
     const Array*, const void*, bool, SerializationType, Buffer*) {
   return LOG_STATUS(Status::SerializationError(
