@@ -5,7 +5,7 @@
  *
  * The MIT License
  *
- * @copyright Copyright (c) 2018-2019 TileDB, Inc.
+ * @copyright Copyright (c) 2018-2021 TileDB, Inc.
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
  * of this software and associated documentation files (the "Software"), to deal
@@ -30,19 +30,36 @@
  * This file defines the info command.
  */
 
-#include <fstream>
-#include <sstream>
-
 #include "commands/info_command.h"
 #include "misc/common.h"
 
+#include "tiledb/common/logger.h"
+#include "tiledb/sm/array/array.h"
+#include "tiledb/sm/array_schema/array_schema.h"
+#include "tiledb/sm/array_schema/attribute.h"
+#include "tiledb/sm/array_schema/dimension.h"
+#include "tiledb/sm/crypto/encryption_key.h"
+#include "tiledb/sm/enums/encryption_type.h"
+#include "tiledb/sm/enums/query_type.h"
+#include "tiledb/sm/fragment/fragment_metadata.h"
 #include "tiledb/sm/storage_manager/storage_manager.h"
 
+#include <cassert>
+#include <fstream>
+#include <iostream>
+#include <sstream>
+
 namespace tiledb {
 namespace cli {
 
 using namespace tiledb::sm;
 
+/** The thread pool for compute-bound tasks. */
+ThreadPool compute_tp_;
+
+/** The thread pool for io-bound tasks. */
+ThreadPool io_tp_;
+
 clipp::group InfoCommand::get_cli() {
   using namespace clipp;
   auto array_arg =
@@ -80,6 +97,9 @@ clipp::group InfoCommand::get_cli() {
 }
 
 void InfoCommand::run() {
+  io_tp_.init(std::thread::hardware_concurrency());
+  compute_tp_.init(std::thread::hardware_concurrency());
+
   switch (type_) {
     case InfoType::None:
       break;
@@ -99,7 +119,8 @@ void InfoCommand::run() {
 }
 
 void InfoCommand::print_tile_sizes() const {
-  StorageManager sm;
+  stats::Stats stats("");
+  StorageManager sm(&compute_tp_, &io_tp_, &stats, tdb_make_shared(Logger, ""));
   THROW_NOT_OK(sm.init(nullptr));
 
   // Open the array
@@ -110,7 +131,7 @@ void InfoCommand::print_tile_sizes() const {
   EncryptionKey enc_key;
 
   // Compute and report mean persisted tile sizes over all attributes.
-  const auto* schema = array.array_schema();
+  const auto* schema = array.array_schema_latest();
   auto fragment_metadata = array.fragment_metadata();
   auto attributes = schema->attributes();
   uint64_t total_persisted_size = 0, total_in_memory_size = 0;
@@ -121,18 +142,20 @@ void InfoCommand::print_tile_sizes() const {
     uint64_t num_tiles = 0;
     for (const auto& f : fragment_metadata) {
       uint64_t tile_num = f->tile_num();
+      std::vector<std::string> names;
+      names.push_back(name);
+      THROW_NOT_OK(f->load_tile_offsets(enc_key, std::move(names)));
+      THROW_NOT_OK(f->load_tile_var_sizes(enc_key, name));
       for (uint64_t tile_idx = 0; tile_idx < tile_num; tile_idx++) {
         uint64_t tile_size = 0;
-        THROW_NOT_OK(
-            f->persisted_tile_size(enc_key, name, tile_idx, &tile_size));
+        THROW_NOT_OK(f->persisted_tile_size(name, tile_idx, &tile_size));
         persisted_tile_size += tile_size;
         in_memory_tile_size += f->tile_size(name, tile_idx);
         num_tiles++;
         if (var_size) {
-          THROW_NOT_OK(
-              f->persisted_tile_var_size(enc_key, name, tile_idx, &tile_size));
+          THROW_NOT_OK(f->persisted_tile_var_size(name, tile_idx, &tile_size));
           persisted_tile_size += tile_size;
-          THROW_NOT_OK(f->tile_var_size(enc_key, name, tile_idx, &tile_size));
+          THROW_NOT_OK(f->tile_var_size(name, tile_idx, &tile_size));
           in_memory_tile_size += tile_size;
           num_tiles++;
         }
@@ -170,7 +193,8 @@ void InfoCommand::print_tile_sizes() const {
 }
 
 void InfoCommand::print_schema_info() const {
-  StorageManager sm;
+  stats::Stats stats("");
+  StorageManager sm(&compute_tp_, &io_tp_, &stats, tdb_make_shared(Logger, ""));
   THROW_NOT_OK(sm.init(nullptr));
 
   // Open the array
@@ -179,14 +203,15 @@ void InfoCommand::print_schema_info() const {
   THROW_NOT_OK(
       array.open(QueryType::READ, EncryptionType::NO_ENCRYPTION, nullptr, 0));
 
-  array.array_schema()->dump(stdout);
+  array.array_schema_latest()->dump(stdout);
 
   // Close the array.
   THROW_NOT_OK(array.close());
 }
 
 void InfoCommand::write_svg_mbrs() const {
-  StorageManager sm;
+  stats::Stats stats("");
+  StorageManager sm(&compute_tp_, &io_tp_, &stats, tdb_make_shared(Logger, ""));
   THROW_NOT_OK(sm.init(nullptr));
 
   // Open the array
@@ -195,7 +220,7 @@ void InfoCommand::write_svg_mbrs() const {
   THROW_NOT_OK(
       array.open(QueryType::READ, EncryptionType::NO_ENCRYPTION, nullptr, 0));
 
-  const auto* schema = array.array_schema();
+  const auto* schema = array.array_schema_latest();
   auto dim_num = schema->dim_num();
   if (dim_num < 2) {
     THROW_NOT_OK(array.close());
@@ -209,9 +234,9 @@ void InfoCommand::write_svg_mbrs() const {
          max_y = std::numeric_limits<double>::min();
   auto fragment_metadata = array.fragment_metadata();
   for (const auto& f : fragment_metadata) {
-    auto mbrs = f->mbrs();
-    for (void* mbr : mbrs) {
-      auto tup = get_mbr(mbr, schema->coords_type());
+    const auto& mbrs = f->mbrs();
+    for (const auto& mbr : mbrs) {
+      auto tup = get_mbr(mbr, schema->domain());
       min_x = std::min(min_x, std::get<0>(tup));
       min_y = std::min(min_y, std::get<1>(tup));
       max_x = std::max(max_x, std::get<0>(tup) + std::get<2>(tup));
@@ -260,7 +285,8 @@ void InfoCommand::write_svg_mbrs() const {
 }
 
 void InfoCommand::write_text_mbrs() const {
-  StorageManager sm;
+  stats::Stats stats("");
+  StorageManager sm(&compute_tp_, &io_tp_, &stats, tdb_make_shared(Logger, ""));
   THROW_NOT_OK(sm.init(nullptr));
 
   // Open the array
@@ -269,15 +295,16 @@ void InfoCommand::write_text_mbrs() const {
   THROW_NOT_OK(
       array.open(QueryType::READ, EncryptionType::NO_ENCRYPTION, nullptr, 0));
 
-  const auto* schema = array.array_schema();
+  auto encryption_key = array.encryption_key();
+  const auto* schema = array.array_schema_latest();
   auto dim_num = schema->dim_num();
   auto fragment_metadata = array.fragment_metadata();
-  auto coords_type = schema->coords_type();
   std::stringstream text;
   for (const auto& f : fragment_metadata) {
-    auto mbrs = f->mbrs();
-    for (void* mbr : mbrs) {
-      auto str_mbr = mbr_to_string(mbr, coords_type, dim_num);
+    f->load_rtree(*encryption_key);
+    const auto& mbrs = f->mbrs();
+    for (const auto& mbr : mbrs) {
+      auto str_mbr = mbr_to_string(mbr, schema->domain());
       for (unsigned i = 0; i < dim_num; i++) {
         text << str_mbr[2 * i + 0] << "," << str_mbr[2 * i + 1];
         if (i < dim_num - 1)
@@ -299,28 +326,53 @@ void InfoCommand::write_text_mbrs() const {
 }
 
 std::tuple<double, double, double, double> InfoCommand::get_mbr(
-    const void* mbr, tiledb::sm::Datatype datatype) const {
-  switch (datatype) {
+    const NDRange& mbr, const Domain* domain) const {
+  assert(domain->dim_num() == 2);
+  double x, y, width, height;
+
+  // First dimension
+  auto d1_type = domain->dimension(0)->type();
+  switch (d1_type) {
     case Datatype::INT8:
-      return get_mbr<int8_t>(mbr);
+      y = static_cast<const int8_t*>(mbr[0].data())[0];
+      height = static_cast<const int8_t*>(mbr[0].data())[1] - y + 1;
+      break;
     case Datatype::UINT8:
-      return get_mbr<uint8_t>(mbr);
+      y = static_cast<const uint8_t*>(mbr[0].data())[0];
+      height = static_cast<const uint8_t*>(mbr[0].data())[1] - y + 1;
+      break;
     case Datatype::INT16:
-      return get_mbr<int16_t>(mbr);
+      y = static_cast<const int16_t*>(mbr[0].data())[0];
+      height = static_cast<const int16_t*>(mbr[0].data())[1] - y + 1;
+      break;
     case Datatype::UINT16:
-      return get_mbr<uint16_t>(mbr);
+      y = static_cast<const uint16_t*>(mbr[0].data())[0];
+      height = static_cast<const uint16_t*>(mbr[0].data())[1] - y + 1;
+      break;
     case Datatype::INT32:
-      return get_mbr<int>(mbr);
+      y = static_cast<const int32_t*>(mbr[0].data())[0];
+      height = static_cast<const int32_t*>(mbr[0].data())[1] - y + 1;
+      break;
     case Datatype::UINT32:
-      return get_mbr<unsigned>(mbr);
+      y = static_cast<const uint32_t*>(mbr[0].data())[0];
+      height = static_cast<const uint32_t*>(mbr[0].data())[1] - y + 1;
+      break;
     case Datatype::INT64:
-      return get_mbr<int64_t>(mbr);
+      y = static_cast<const int64_t*>(mbr[0].data())[0];
+      height = static_cast<const int64_t*>(mbr[0].data())[1] - y + 1;
+      break;
     case Datatype::UINT64:
-      return get_mbr<uint64_t>(mbr);
+      y = static_cast<const uint64_t*>(mbr[0].data())[0];
+      height = static_cast<const uint64_t*>(mbr[0].data())[1] - y + 1;
+      break;
     case Datatype::FLOAT32:
-      return get_mbr<float>(mbr);
+      y = static_cast<const float*>(mbr[0].data())[0];
+      height = static_cast<const float*>(mbr[0].data())[1] - y + 1;
+      break;
     case Datatype::FLOAT64:
-      return get_mbr<double>(mbr);
+      y = static_cast<const double*>(mbr[0].data())[0];
+      height = static_cast<const double*>(mbr[0].data())[1] - y + 1;
+      break;
     case Datatype::DATETIME_YEAR:
     case Datatype::DATETIME_MONTH:
     case Datatype::DATETIME_WEEK:
@@ -334,88 +386,170 @@ std::tuple<double, double, double, double> InfoCommand::get_mbr(
     case Datatype::DATETIME_PS:
     case Datatype::DATETIME_FS:
     case Datatype::DATETIME_AS:
-      return get_mbr<int64_t>(mbr);
+    case Datatype::TIME_HR:
+    case Datatype::TIME_MIN:
+    case Datatype::TIME_SEC:
+    case Datatype::TIME_MS:
+    case Datatype::TIME_US:
+    case Datatype::TIME_NS:
+    case Datatype::TIME_PS:
+    case Datatype::TIME_FS:
+    case Datatype::TIME_AS:
+      y = static_cast<const int64_t*>(mbr[0].data())[0];
+      height = static_cast<const int64_t*>(mbr[0].data())[1] - y + 1;
+      break;
+    default:
+      throw std::invalid_argument(
+          "Cannot get MBR; Unsupported coordinates type");
+  }
+
+  // Second dimension
+  auto d2_type = domain->dimension(1)->type();
+  switch (d2_type) {
+    case Datatype::INT8:
+      x = static_cast<const int8_t*>(mbr[1].data())[0];
+      width = static_cast<const int8_t*>(mbr[1].data())[1] - x + 1;
+      break;
+    case Datatype::UINT8:
+      x = static_cast<const uint8_t*>(mbr[1].data())[0];
+      width = static_cast<const uint8_t*>(mbr[1].data())[1] - x + 1;
+      break;
+    case Datatype::INT16:
+      x = static_cast<const int16_t*>(mbr[1].data())[0];
+      width = static_cast<const int16_t*>(mbr[1].data())[1] - x + 1;
+      break;
+    case Datatype::UINT16:
+      x = static_cast<const uint16_t*>(mbr[1].data())[0];
+      width = static_cast<const uint16_t*>(mbr[1].data())[1] - x + 1;
+      break;
+    case Datatype::INT32:
+      x = static_cast<const int32_t*>(mbr[1].data())[0];
+      width = static_cast<const int32_t*>(mbr[1].data())[1] - x + 1;
+      break;
+    case Datatype::UINT32:
+      x = static_cast<const uint32_t*>(mbr[1].data())[0];
+      width = static_cast<const uint32_t*>(mbr[1].data())[1] - x + 1;
+      break;
+    case Datatype::INT64:
+      x = static_cast<const int64_t*>(mbr[1].data())[0];
+      width = static_cast<const int64_t*>(mbr[1].data())[1] - x + 1;
+      break;
+    case Datatype::UINT64:
+      x = static_cast<const uint64_t*>(mbr[1].data())[0];
+      width = static_cast<const uint64_t*>(mbr[1].data())[1] - x + 1;
+      break;
+    case Datatype::FLOAT32:
+      x = static_cast<const float*>(mbr[1].data())[0];
+      width = static_cast<const float*>(mbr[1].data())[1] - x + 1;
+      break;
+    case Datatype::FLOAT64:
+      x = static_cast<const double*>(mbr[1].data())[0];
+      width = static_cast<const double*>(mbr[1].data())[1] - x + 1;
+      break;
+    case Datatype::DATETIME_YEAR:
+    case Datatype::DATETIME_MONTH:
+    case Datatype::DATETIME_WEEK:
+    case Datatype::DATETIME_DAY:
+    case Datatype::DATETIME_HR:
+    case Datatype::DATETIME_MIN:
+    case Datatype::DATETIME_SEC:
+    case Datatype::DATETIME_MS:
+    case Datatype::DATETIME_US:
+    case Datatype::DATETIME_NS:
+    case Datatype::DATETIME_PS:
+    case Datatype::DATETIME_FS:
+    case Datatype::DATETIME_AS:
+    case Datatype::TIME_HR:
+    case Datatype::TIME_MIN:
+    case Datatype::TIME_SEC:
+    case Datatype::TIME_MS:
+    case Datatype::TIME_US:
+    case Datatype::TIME_NS:
+    case Datatype::TIME_PS:
+    case Datatype::TIME_FS:
+    case Datatype::TIME_AS:
+      x = static_cast<const int64_t*>(mbr[1].data())[0];
+      width = static_cast<const int64_t*>(mbr[1].data())[1] - x + 1;
+      break;
     default:
       throw std::invalid_argument(
           "Cannot get MBR; Unsupported coordinates type");
   }
-}
 
-template <typename T>
-std::tuple<double, double, double, double> InfoCommand::get_mbr(
-    const void* mbr) const {
-  T x, y, width, height;
-  y = static_cast<const T*>(mbr)[0];
-  height = static_cast<const T*>(mbr)[1] - y + 1;
-  x = static_cast<const T*>(mbr)[2];
-  width = static_cast<const T*>(mbr)[3] - x + 1;
   return std::make_tuple(x, y, width, height);
 }
 
+// Works only for fixed-sized coordinates
 std::vector<std::string> InfoCommand::mbr_to_string(
-    const void* mbr, Datatype coords_type, unsigned dim_num) const {
+    const NDRange& mbr, const Domain* domain) const {
   std::vector<std::string> result;
-  for (unsigned i = 0; i < dim_num; i++) {
-    switch (coords_type) {
+  const int8_t* r8;
+  const uint8_t* ru8;
+  const int16_t* r16;
+  const uint16_t* ru16;
+  const int32_t* r32;
+  const uint32_t* ru32;
+  const int64_t* r64;
+  const uint64_t* ru64;
+  const float* rf32;
+  const double* rf64;
+  auto dim_num = domain->dim_num();
+  for (unsigned d = 0; d < dim_num; d++) {
+    auto type = domain->dimension(d)->type();
+    switch (type) {
+      case sm::Datatype::STRING_ASCII:
+        result.push_back(mbr[d].start_str());
+        result.push_back(mbr[d].end_str());
+        break;
       case Datatype::INT8:
-        result.push_back(
-            std::to_string(static_cast<const uint8_t*>(mbr)[2 * i + 0]));
-        result.push_back(
-            std::to_string(static_cast<const uint8_t*>(mbr)[2 * i + 1]));
+        r8 = (const int8_t*)mbr[d].data();
+        result.push_back(std::to_string(r8[0]));
+        result.push_back(std::to_string(r8[1]));
         break;
       case Datatype::UINT8:
-        result.push_back(
-            std::to_string(static_cast<const uint8_t*>(mbr)[2 * i + 0]));
-        result.push_back(
-            std::to_string(static_cast<const uint8_t*>(mbr)[2 * i + 1]));
+        ru8 = (const uint8_t*)mbr[d].data();
+        result.push_back(std::to_string(ru8[0]));
+        result.push_back(std::to_string(ru8[1]));
         break;
       case Datatype::INT16:
-        result.push_back(
-            std::to_string(static_cast<const int16_t*>(mbr)[2 * i + 0]));
-        result.push_back(
-            std::to_string(static_cast<const int16_t*>(mbr)[2 * i + 1]));
+        r16 = (const int16_t*)mbr[d].data();
+        result.push_back(std::to_string(r16[0]));
+        result.push_back(std::to_string(r16[1]));
         break;
       case Datatype::UINT16:
-        result.push_back(
-            std::to_string(static_cast<const uint16_t*>(mbr)[2 * i + 0]));
-        result.push_back(
-            std::to_string(static_cast<const uint16_t*>(mbr)[2 * i + 1]));
+        ru16 = (const uint16_t*)mbr[d].data();
+        result.push_back(std::to_string(ru16[0]));
+        result.push_back(std::to_string(ru16[1]));
         break;
       case Datatype::INT32:
-        result.push_back(
-            std::to_string(static_cast<const int*>(mbr)[2 * i + 0]));
-        result.push_back(
-            std::to_string(static_cast<const int*>(mbr)[2 * i + 1]));
+        r32 = (const int32_t*)mbr[d].data();
+        result.push_back(std::to_string(r32[0]));
+        result.push_back(std::to_string(r32[1]));
         break;
       case Datatype::UINT32:
-        result.push_back(
-            std::to_string(static_cast<const unsigned*>(mbr)[2 * i + 0]));
-        result.push_back(
-            std::to_string(static_cast<const unsigned*>(mbr)[2 * i + 1]));
+        ru32 = (const uint32_t*)mbr[d].data();
+        result.push_back(std::to_string(ru32[0]));
+        result.push_back(std::to_string(ru32[1]));
         break;
       case Datatype::INT64:
-        result.push_back(
-            std::to_string(static_cast<const int64_t*>(mbr)[2 * i + 0]));
-        result.push_back(
-            std::to_string(static_cast<const int64_t*>(mbr)[2 * i + 1]));
+        r64 = (const int64_t*)mbr[d].data();
+        result.push_back(std::to_string(r64[0]));
+        result.push_back(std::to_string(r64[1]));
         break;
       case Datatype::UINT64:
-        result.push_back(
-            std::to_string(static_cast<const uint64_t*>(mbr)[2 * i + 0]));
-        result.push_back(
-            std::to_string(static_cast<const uint64_t*>(mbr)[2 * i + 1]));
+        ru64 = (const uint64_t*)mbr[d].data();
+        result.push_back(std::to_string(ru64[0]));
+        result.push_back(std::to_string(ru64[1]));
         break;
       case Datatype::FLOAT32:
-        result.push_back(
-            std::to_string(static_cast<const float*>(mbr)[2 * i + 0]));
-        result.push_back(
-            std::to_string(static_cast<const float*>(mbr)[2 * i + 1]));
+        rf32 = (const float*)mbr[d].data();
+        result.push_back(std::to_string(rf32[0]));
+        result.push_back(std::to_string(rf32[1]));
         break;
       case Datatype::FLOAT64:
-        result.push_back(
-            std::to_string(static_cast<const double*>(mbr)[2 * i + 0]));
-        result.push_back(
-            std::to_string(static_cast<const double*>(mbr)[2 * i + 1]));
+        rf64 = (const double*)mbr[d].data();
+        result.push_back(std::to_string(rf64[0]));
+        result.push_back(std::to_string(rf64[1]));
         break;
       default:
         throw std::invalid_argument(
@@ -426,27 +560,5 @@ std::vector<std::string> InfoCommand::mbr_to_string(
   return result;
 }
 
-// Explicit template instantiations
-template std::tuple<double, double, double, double>
-InfoCommand::get_mbr<int8_t>(const void* mbr) const;
-template std::tuple<double, double, double, double>
-InfoCommand::get_mbr<uint8_t>(const void* mbr) const;
-template std::tuple<double, double, double, double>
-InfoCommand::get_mbr<int16_t>(const void* mbr) const;
-template std::tuple<double, double, double, double>
-InfoCommand::get_mbr<uint16_t>(const void* mbr) const;
-template std::tuple<double, double, double, double>
-InfoCommand::get_mbr<int32_t>(const void* mbr) const;
-template std::tuple<double, double, double, double>
-InfoCommand::get_mbr<uint32_t>(const void* mbr) const;
-template std::tuple<double, double, double, double>
-InfoCommand::get_mbr<int64_t>(const void* mbr) const;
-template std::tuple<double, double, double, double>
-InfoCommand::get_mbr<uint64_t>(const void* mbr) const;
-template std::tuple<double, double, double, double> InfoCommand::get_mbr<float>(
-    const void* mbr) const;
-template std::tuple<double, double, double, double>
-InfoCommand::get_mbr<double>(const void* mbr) const;
-
 }  // namespace cli
 }  // namespace tiledb
